import os
import time
import ssl
import logging
import pkg_resources
import google.protobuf.text_format
from . import abi
from . import casper_pb2 as casper
from . import consensus_pb2 as consensus
from . import crypto


def _read_binary(file_name: str):
    with open(file_name, "rb") as f:
        return f.read()


def hexify(o):
    """
    Convert protobuf message to text format with cryptographic keys and signatures in base 16.
    """
    return google.protobuf.text_format.MessageToString(o)


def bundled_contract(file_name):
    """
    Return path to contract file bundled with the package.
    """
    p = pkg_resources.resource_filename(__name__, file_name)
    if not os.path.exists(p):
        raise Exception(f"Missing bundled contract {file_name} ({p})")
    return p


def extract_common_name(certificate_file: str) -> str:
    cert_dict = ssl._ssl._test_decode_cert(certificate_file)
    return [t[0][1] for t in cert_dict["subject"] if t[0][0] == "commonName"][0]


# Note, there is also casper.StateQuery.KeyVariant.KEY_VARIANT_UNSPECIFIED,
# but it doesn't seem to have an official string representation
# ("key_variant_unspecified"? "unspecified"?) and is not used by the client.
STATE_QUERY_KEY_VARIANT = {
    "hash": casper.StateQuery.KeyVariant.HASH,
    "uref": casper.StateQuery.KeyVariant.UREF,
    "address": casper.StateQuery.KeyVariant.ADDRESS,
    "local": casper.StateQuery.KeyVariant.LOCAL,
}


def key_variant(key_type):
    """
    Returns query-state key variant.
    """
    variant = STATE_QUERY_KEY_VARIANT.get(key_type.lower(), None)
    if variant is None:
        raise Exception(f"{key_type} is not a known query-state key type")
    return variant


def _encode_contract(contract_options, contract_args):
    file_name, hash, name, uref = contract_options
    C = consensus.Deploy.Code
    if file_name:
        return C(wasm=_read_binary(file_name), args=contract_args)
    if hash:
        return C(hash=hash, args=contract_args)
    if name:
        return C(name=name, args=contract_args)
    if uref:
        return C(uref=uref, args=contract_args)
    raise Exception("One of wasm, hash, name or uref is required")


def _serialize(o) -> bytes:
    return o.SerializeToString()


def make_deploy(
    from_addr: bytes = None,
    gas_price: int = 10,
    payment: str = None,
    session: str = None,
    public_key: str = None,
    session_args: bytes = None,
    payment_args: bytes = None,
    payment_amount: int = None,
    payment_hash: bytes = None,
    payment_name: str = None,
    payment_uref: bytes = None,
    session_hash: bytes = None,
    session_name: str = None,
    session_uref: bytes = None,
    ttl_millis: int = 0,
    dependencies: list = None,
    chain_name: str = None,
):
    """
    Create a protobuf deploy object. See deploy for description of parameters.
    """
    # Convert from hex to binary.
    if from_addr and len(from_addr) == 64:
        from_addr = bytes.fromhex(from_addr)

    if from_addr and len(from_addr) != 32:
        raise Exception(f"from_addr must be 32 bytes")

    if payment_amount:
        payment_args = abi.ABI.args([abi.ABI.big_int("amount", int(payment_amount))])

    # Unless one of payment* options supplied use bundled standard-payment
    if not any((payment, payment_name, payment_hash, payment_uref)):
        payment = bundled_contract("standard_payment.wasm")

    session_options = (session, session_hash, session_name, session_uref)
    payment_options = (payment, payment_hash, payment_name, payment_uref)

    # Compatibility mode, should be removed when payment is obligatory
    if len(list(filter(None, payment_options))) == 0:
        logging.info("No payment contract provided, using session as payment")
        payment_options = session_options

    if len(list(filter(None, session_options))) != 1:
        raise TypeError(
            "deploy: only one of session, session_hash, session_name, session_uref must be provided"
        )

    if len(list(filter(None, payment_options))) != 1:
        raise TypeError(
            "deploy: only one of payment, payment_hash, payment_name, payment_uref must be provided"
        )

    # session_args must go to payment as well for now cause otherwise we'll get GASLIMIT error,
    # if payment is same as session:
    # https://github.com/CasperLabs/CasperLabs/blob/dev/casper/src/main/scala/io/casperlabs/casper/util/ProtoUtil.scala#L463
    body = consensus.Deploy.Body(
        session=_encode_contract(session_options, session_args),
        payment=_encode_contract(payment_options, payment_args),
    )

    header = consensus.Deploy.Header(
        account_public_key=from_addr
        or (public_key and crypto.read_pem_key(public_key)),
        timestamp=int(1000 * time.time()),
        gas_price=gas_price,
        body_hash=crypto.blake2b_hash(_serialize(body)),
        ttl_millis=ttl_millis,
        dependencies=dependencies and [bytes.fromhex(d) for d in dependencies] or [],
        chain_name=chain_name or "",
    )

    deploy_hash = crypto.blake2b_hash(_serialize(header))

    return consensus.Deploy(deploy_hash=deploy_hash, header=header, body=body)


def sign_deploy(deploy, public_key, private_key_file):
    # See if this is hex encoded
    try:
        public_key = bytes.fromhex(public_key)
    except TypeError:
        pass

    deploy.approvals.extend(
        [
            consensus.Approval(
                approver_public_key=public_key,
                signature=crypto.signature(private_key_file, deploy.deploy_hash),
            )
        ]
    )
    return deploy


def get_public_key(public_key, from_addr, private_key):
    return (
        (public_key and crypto.read_pem_key(public_key))
        or from_addr
        or crypto.private_to_public_key(private_key)
    )



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import ipc_pb2 as ipc__pb2


class ExecutionEngineServiceStub(object):
  """--- END PROOF-OF-STAKE SERVICE DEFINITION --- //

  Definition of the service.
  ExecutionEngine implements server part while Consensus implements client part.
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.commit = channel.unary_unary(
        '/io.casperlabs.ipc.ExecutionEngineService/commit',
        request_serializer=ipc__pb2.CommitRequest.SerializeToString,
        response_deserializer=ipc__pb2.CommitResponse.FromString,
        )
    self.query = channel.unary_unary(
        '/io.casperlabs.ipc.ExecutionEngineService/query',
        request_serializer=ipc__pb2.QueryRequest.SerializeToString,
        response_deserializer=ipc__pb2.QueryResponse.FromString,
        )
    self.execute = channel.unary_unary(
        '/io.casperlabs.ipc.ExecutionEngineService/execute',
        request_serializer=ipc__pb2.ExecuteRequest.SerializeToString,
        response_deserializer=ipc__pb2.ExecuteResponse.FromString,
        )
    self.run_genesis = channel.unary_unary(
        '/io.casperlabs.ipc.ExecutionEngineService/run_genesis',
        request_serializer=ipc__pb2.ChainSpec.GenesisConfig.SerializeToString,
        response_deserializer=ipc__pb2.GenesisResponse.FromString,
        )
    self.upgrade = channel.unary_unary(
        '/io.casperlabs.ipc.ExecutionEngineService/upgrade',
        request_serializer=ipc__pb2.UpgradeRequest.SerializeToString,
        response_deserializer=ipc__pb2.UpgradeResponse.FromString,
        )
    self.bid_state = channel.unary_unary(
        '/io.casperlabs.ipc.ExecutionEngineService/bid_state',
        request_serializer=ipc__pb2.BidStateRequest.SerializeToString,
        response_deserializer=ipc__pb2.BidStateResponse.FromString,
        )
    self.distribute_rewards = channel.unary_unary(
        '/io.casperlabs.ipc.ExecutionEngineService/distribute_rewards',
        request_serializer=ipc__pb2.DistributeRewardsRequest.SerializeToString,
        response_deserializer=ipc__pb2.DistributeRewardsResponse.FromString,
        )
    self.slash = channel.unary_unary(
        '/io.casperlabs.ipc.ExecutionEngineService/slash',
        request_serializer=ipc__pb2.SlashRequest.SerializeToString,
        response_deserializer=ipc__pb2.SlashResponse.FromString,
        )
    self.unbond_payout = channel.unary_unary(
        '/io.casperlabs.ipc.ExecutionEngineService/unbond_payout',
        request_serializer=ipc__pb2.UnbondPayoutRequest.SerializeToString,
        response_deserializer=ipc__pb2.UnbondPayoutResponse.FromString,
        )


class ExecutionEngineServiceServicer(object):
  """--- END PROOF-OF-STAKE SERVICE DEFINITION --- //

  Definition of the service.
  ExecutionEngine implements server part while Consensus implements client part.
  """

  def commit(self, request, context):
    """execution endpoints
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def query(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def execute(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def run_genesis(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def upgrade(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def bid_state(self, request, context):
    """proof-of-stake endpoints
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def distribute_rewards(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def slash(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def unbond_payout(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_ExecutionEngineServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'commit': grpc.unary_unary_rpc_method_handler(
          servicer.commit,
          request_deserializer=ipc__pb2.CommitRequest.FromString,
          response_serializer=ipc__pb2.CommitResponse.SerializeToString,
      ),
      'query': grpc.unary_unary_rpc_method_handler(
          servicer.query,
          request_deserializer=ipc__pb2.QueryRequest.FromString,
          response_serializer=ipc__pb2.QueryResponse.SerializeToString,
      ),
      'execute': grpc.unary_unary_rpc_method_handler(
          servicer.execute,
          request_deserializer=ipc__pb2.ExecuteRequest.FromString,
          response_serializer=ipc__pb2.ExecuteResponse.SerializeToString,
      ),
      'run_genesis': grpc.unary_unary_rpc_method_handler(
          servicer.run_genesis,
          request_deserializer=ipc__pb2.ChainSpec.GenesisConfig.FromString,
          response_serializer=ipc__pb2.GenesisResponse.SerializeToString,
      ),
      'upgrade': grpc.unary_unary_rpc_method_handler(
          servicer.upgrade,
          request_deserializer=ipc__pb2.UpgradeRequest.FromString,
          response_serializer=ipc__pb2.UpgradeResponse.SerializeToString,
      ),
      'bid_state': grpc.unary_unary_rpc_method_handler(
          servicer.bid_state,
          request_deserializer=ipc__pb2.BidStateRequest.FromString,
          response_serializer=ipc__pb2.BidStateResponse.SerializeToString,
      ),
      'distribute_rewards': grpc.unary_unary_rpc_method_handler(
          servicer.distribute_rewards,
          request_deserializer=ipc__pb2.DistributeRewardsRequest.FromString,
          response_serializer=ipc__pb2.DistributeRewardsResponse.SerializeToString,
      ),
      'slash': grpc.unary_unary_rpc_method_handler(
          servicer.slash,
          request_deserializer=ipc__pb2.SlashRequest.FromString,
          response_serializer=ipc__pb2.SlashResponse.SerializeToString,
      ),
      'unbond_payout': grpc.unary_unary_rpc_method_handler(
          servicer.unbond_payout,
          request_deserializer=ipc__pb2.UnbondPayoutRequest.FromString,
          response_serializer=ipc__pb2.UnbondPayoutResponse.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'io.casperlabs.ipc.ExecutionEngineService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: storage.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import consensus_pb2 as consensus__pb2
from . import transforms_pb2 as transforms__pb2
from . import scalapb_pb2 as scalapb__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='storage.proto',
  package='io.casperlabs.storage',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=b'\n\rstorage.proto\x12\x15io.casperlabs.storage\x1a\x0f\x63onsensus.proto\x1a\x10transforms.proto\x1a\rscalapb.proto\"\x80\x02\n\x15\x42lockMsgWithTransform\x12<\n\rblock_message\x18\x01 \x01(\x0b\x32%.io.casperlabs.casper.consensus.Block\x12P\n\rblock_effects\x18\x03 \x03(\x0b\x32\x39.io.casperlabs.storage.BlockMsgWithTransform.StageEffects\x1aQ\n\x0cStageEffects\x12\r\n\x05stage\x18\x01 \x01(\r\x12\x32\n\x07\x65\x66\x66\x65\x63ts\x18\x02 \x03(\x0b\x32!.io.casperlabs.ipc.TransformEntryJ\x04\x08\x02\x10\x03\"\x93\x03\n\x15\x42lockMetadataInternal\x12\x11\n\tblockHash\x18\x01 \x01(\x0c\x12/\n\x07parents\x18\x02 \x03(\x0c\x42\x1e\xe2?\x1b\x1a\x19\x63ollection.immutable.List\x12\x1c\n\x14validator_public_key\x18\x03 \x01(\x0c\x12k\n\x0ejustifications\x18\x04 \x03(\x0b\x32\x33.io.casperlabs.casper.consensus.Block.JustificationB\x1e\xe2?\x1b\x1a\x19\x63ollection.immutable.List\x12S\n\x05\x62onds\x18\x05 \x03(\x0b\x32$.io.casperlabs.casper.consensus.BondB\x1e\xe2?\x1b\x1a\x19\x63ollection.immutable.List\x12\x0c\n\x04rank\x18\x06 \x01(\x03\x12\x1f\n\x17validator_block_seq_num\x18\x07 \x01(\x05:\'\xe2?$\"\"io.casperlabs.models.BlockMetadatab\x06proto3'
  ,
  dependencies=[consensus__pb2.DESCRIPTOR,transforms__pb2.DESCRIPTOR,scalapb__pb2.DESCRIPTOR,])




_BLOCKMSGWITHTRANSFORM_STAGEEFFECTS = _descriptor.Descriptor(
  name='StageEffects',
  full_name='io.casperlabs.storage.BlockMsgWithTransform.StageEffects',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='stage', full_name='io.casperlabs.storage.BlockMsgWithTransform.StageEffects.stage', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='effects', full_name='io.casperlabs.storage.BlockMsgWithTransform.StageEffects.effects', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=260,
  serialized_end=341,
)

_BLOCKMSGWITHTRANSFORM = _descriptor.Descriptor(
  name='BlockMsgWithTransform',
  full_name='io.casperlabs.storage.BlockMsgWithTransform',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='block_message', full_name='io.casperlabs.storage.BlockMsgWithTransform.block_message', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_effects', full_name='io.casperlabs.storage.BlockMsgWithTransform.block_effects', index=1,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_BLOCKMSGWITHTRANSFORM_STAGEEFFECTS, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=91,
  serialized_end=347,
)


_BLOCKMETADATAINTERNAL = _descriptor.Descriptor(
  name='BlockMetadataInternal',
  full_name='io.casperlabs.storage.BlockMetadataInternal',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='blockHash', full_name='io.casperlabs.storage.BlockMetadataInternal.blockHash', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='parents', full_name='io.casperlabs.storage.BlockMetadataInternal.parents', index=1,
      number=2, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\342?\033\032\031collection.immutable.List', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='validator_public_key', full_name='io.casperlabs.storage.BlockMetadataInternal.validator_public_key', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='justifications', full_name='io.casperlabs.storage.BlockMetadataInternal.justifications', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\342?\033\032\031collection.immutable.List', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bonds', full_name='io.casperlabs.storage.BlockMetadataInternal.bonds', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\342?\033\032\031collection.immutable.List', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='rank', full_name='io.casperlabs.storage.BlockMetadataInternal.rank', index=5,
      number=6, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='validator_block_seq_num', full_name='io.casperlabs.storage.BlockMetadataInternal.validator_block_seq_num', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'\342?$\"\"io.casperlabs.models.BlockMetadata',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=350,
  serialized_end=753,
)

_BLOCKMSGWITHTRANSFORM_STAGEEFFECTS.fields_by_name['effects'].message_type = transforms__pb2._TRANSFORMENTRY
_BLOCKMSGWITHTRANSFORM_STAGEEFFECTS.containing_type = _BLOCKMSGWITHTRANSFORM
_BLOCKMSGWITHTRANSFORM.fields_by_name['block_message'].message_type = consensus__pb2._BLOCK
_BLOCKMSGWITHTRANSFORM.fields_by_name['block_effects'].message_type = _BLOCKMSGWITHTRANSFORM_STAGEEFFECTS
_BLOCKMETADATAINTERNAL.fields_by_name['justifications'].message_type = consensus__pb2._BLOCK_JUSTIFICATION
_BLOCKMETADATAINTERNAL.fields_by_name['bonds'].message_type = consensus__pb2._BOND
DESCRIPTOR.message_types_by_name['BlockMsgWithTransform'] = _BLOCKMSGWITHTRANSFORM
DESCRIPTOR.message_types_by_name['BlockMetadataInternal'] = _BLOCKMETADATAINTERNAL
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

BlockMsgWithTransform = _reflection.GeneratedProtocolMessageType('BlockMsgWithTransform', (_message.Message,), {

  'StageEffects' : _reflection.GeneratedProtocolMessageType('StageEffects', (_message.Message,), {
    'DESCRIPTOR' : _BLOCKMSGWITHTRANSFORM_STAGEEFFECTS,
    '__module__' : 'storage_pb2'
    # @@protoc_insertion_point(class_scope:io.casperlabs.storage.BlockMsgWithTransform.StageEffects)
    })
  ,
  'DESCRIPTOR' : _BLOCKMSGWITHTRANSFORM,
  '__module__' : 'storage_pb2'
  # @@protoc_insertion_point(class_scope:io.casperlabs.storage.BlockMsgWithTransform)
  })
_sym_db.RegisterMessage(BlockMsgWithTransform)
_sym_db.RegisterMessage(BlockMsgWithTransform.StageEffects)

BlockMetadataInternal = _reflection.GeneratedProtocolMessageType('BlockMetadataInternal', (_message.Message,), {
  'DESCRIPTOR' : _BLOCKMETADATAINTERNAL,
  '__module__' : 'storage_pb2'
  # @@protoc_insertion_point(class_scope:io.casperlabs.storage.BlockMetadataInternal)
  })
_sym_db.RegisterMessage(BlockMetadataInternal)


_BLOCKMETADATAINTERNAL.fields_by_name['parents']._options = None
_BLOCKMETADATAINTERNAL.fields_by_name['justifications']._options = None
_BLOCKMETADATAINTERNAL.fields_by_name['bonds']._options = None
_BLOCKMETADATAINTERNAL._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: node.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='node.proto',
  package='io.casperlabs.comm.discovery',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=b'\n\nnode.proto\x12\x1cio.casperlabs.comm.discovery\"a\n\x04Node\x12\n\n\x02id\x18\x01 \x01(\x0c\x12\x0c\n\x04host\x18\x02 \x01(\t\x12\x15\n\rprotocol_port\x18\x03 \x01(\r\x12\x16\n\x0e\x64iscovery_port\x18\x04 \x01(\r\x12\x10\n\x08\x63hain_id\x18\x05 \x01(\x0c\x62\x06proto3'
)




_NODE = _descriptor.Descriptor(
  name='Node',
  full_name='io.casperlabs.comm.discovery.Node',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='io.casperlabs.comm.discovery.Node.id', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='host', full_name='io.casperlabs.comm.discovery.Node.host', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='protocol_port', full_name='io.casperlabs.comm.discovery.Node.protocol_port', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='discovery_port', full_name='io.casperlabs.comm.discovery.Node.discovery_port', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='chain_id', full_name='io.casperlabs.comm.discovery.Node.chain_id', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=44,
  serialized_end=141,
)

DESCRIPTOR.message_types_by_name['Node'] = _NODE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Node = _reflection.GeneratedProtocolMessageType('Node', (_message.Message,), {
  'DESCRIPTOR' : _NODE,
  '__module__' : 'node_pb2'
  # @@protoc_insertion_point(class_scope:io.casperlabs.comm.discovery.Node)
  })
_sym_db.RegisterMessage(Node)


# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import control_pb2 as control__pb2


class ControlServiceStub(object):
  """The operators can use the ControlService to issue commands to the node,
  such as proposing a block or otherwise influence the proposal strategy.
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.Propose = channel.unary_unary(
        '/io.casperlabs.node.api.control.ControlService/Propose',
        request_serializer=control__pb2.ProposeRequest.SerializeToString,
        response_deserializer=control__pb2.ProposeResponse.FromString,
        )


class ControlServiceServicer(object):
  """The operators can use the ControlService to issue commands to the node,
  such as proposing a block or otherwise influence the proposal strategy.
  """

  def Propose(self, request, context):
    """Propose a block using the deploys in the pool.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_ControlServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'Propose': grpc.unary_unary_rpc_method_handler(
          servicer.Propose,
          request_deserializer=control__pb2.ProposeRequest.FromString,
          response_serializer=control__pb2.ProposeResponse.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'io.casperlabs.node.api.control.ControlService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: control.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import annotations_pb2 as annotations__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='control.proto',
  package='io.casperlabs.node.api.control',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=b'\n\rcontrol.proto\x12\x1eio.casperlabs.node.api.control\x1a\x11\x61nnotations.proto\"\x10\n\x0eProposeRequest\"%\n\x0fProposeResponse\x12\x12\n\nblock_hash\x18\x01 \x01(\x0c\x32\x95\x01\n\x0e\x43ontrolService\x12\x82\x01\n\x07Propose\x12..io.casperlabs.node.api.control.ProposeRequest\x1a/.io.casperlabs.node.api.control.ProposeResponse\"\x16\x82\xd3\xe4\x93\x02\x10\"\x0b/v2:propose:\x01*b\x06proto3'
  ,
  dependencies=[annotations__pb2.DESCRIPTOR,])




_PROPOSEREQUEST = _descriptor.Descriptor(
  name='ProposeRequest',
  full_name='io.casperlabs.node.api.control.ProposeRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=68,
  serialized_end=84,
)


_PROPOSERESPONSE = _descriptor.Descriptor(
  name='ProposeResponse',
  full_name='io.casperlabs.node.api.control.ProposeResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='block_hash', full_name='io.casperlabs.node.api.control.ProposeResponse.block_hash', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=86,
  serialized_end=123,
)

DESCRIPTOR.message_types_by_name['ProposeRequest'] = _PROPOSEREQUEST
DESCRIPTOR.message_types_by_name['ProposeResponse'] = _PROPOSERESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ProposeRequest = _reflection.GeneratedProtocolMessageType('ProposeRequest', (_message.Message,), {
  'DESCRIPTOR' : _PROPOSEREQUEST,
  '__module__' : 'control_pb2'
  # @@protoc_insertion_point(class_scope:io.casperlabs.node.api.control.ProposeRequest)
  })
_sym_db.RegisterMessage(ProposeRequest)

ProposeResponse = _reflection.GeneratedProtocolMessageType('ProposeResponse', (_message.Message,), {
  'DESCRIPTOR' : _PROPOSERESPONSE,
  '__module__' : 'control_pb2'
  # @@protoc_insertion_point(class_scope:io.casperlabs.node.api.control.ProposeResponse)
  })
_sym_db.RegisterMessage(ProposeResponse)



_CONTROLSERVICE = _descriptor.ServiceDescriptor(
  name='ControlService',
  full_name='io.casperlabs.node.api.control.ControlService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=126,
  serialized_end=275,
  methods=[
  _descriptor.MethodDescriptor(
    name='Propose',
    full_name='io.casperlabs.node.api.control.ControlService.Propose',
    index=0,
    containing_service=None,
    input_type=_PROPOSEREQUEST,
    output_type=_PROPOSERESPONSE,
    serialized_options=b'\202\323\344\223\002\020\"\013/v2:propose:\001*',
  ),
])
_sym_db.RegisterServiceDescriptor(_CONTROLSERVICE)

DESCRIPTOR.services_by_name['ControlService'] = _CONTROLSERVICE

# @@protoc_insertion_point(module_scope)



# Generated by the Protocol Buffers compiler. DO NOT EDIT!
# source: control.proto
# plugin: grpclib.plugin.main
import abc
import typing

import grpclib.const
import grpclib.client
if typing.TYPE_CHECKING:
    import grpclib.server

from . import annotations_pb2
from . import control_pb2


class ControlServiceBase(abc.ABC):

    @abc.abstractmethod
    async def Propose(self, stream: 'grpclib.server.Stream[control_pb2.ProposeRequest, control_pb2.ProposeResponse]') -> None:
        pass

    def __mapping__(self) -> typing.Dict[str, grpclib.const.Handler]:
        return {
            '/io.casperlabs.node.api.control.ControlService/Propose': grpclib.const.Handler(
                self.Propose,
                grpclib.const.Cardinality.UNARY_UNARY,
                control_pb2.ProposeRequest,
                control_pb2.ProposeResponse,
            ),
        }


class ControlServiceStub:

    def __init__(self, channel: grpclib.client.Channel) -> None:
        self.Propose = grpclib.client.UnaryUnaryMethod(
            channel,
            '/io.casperlabs.node.api.control.ControlService/Propose',
            control_pb2.ProposeRequest,
            control_pb2.ProposeResponse,
        )



"""
Cryptography related code used in the CasperLabs client.
"""

import datetime
import base64
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric import ed25519 as cryptography_ed25519
from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.x509.oid import NameOID
from Crypto.Hash import keccak
from pyblake2 import blake2b
import ed25519
from . import consensus_pb2 as consensus


def generate_validators_keys():
    validator_private = cryptography_ed25519.Ed25519PrivateKey.generate()
    validator_public = validator_private.public_key()

    validator_private_pem = validator_private.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )

    validator_public_pem = validator_public.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )
    validator_public_bytes = validator_public.public_bytes(
        encoding=serialization.Encoding.Raw, format=serialization.PublicFormat.Raw
    )
    return validator_private_pem, validator_public_pem, validator_public_bytes


def read_pem_key(file_name: str):
    with open(file_name) as f:
        s = [l for l in f.readlines() if l and not l.startswith("-----")][0].strip()
        r = base64.b64decode(s)
        return len(r) % 32 == 0 and r[:32] or r[-32:]


def generate_key_pair():
    curve = ec.SECP256R1()
    private_key = ec.generate_private_key(curve, default_backend())
    public_key = private_key.public_key()
    return private_key, public_key


def public_address(public_key):
    numbers = public_key.public_numbers()
    x, y = numbers.x, numbers.y

    def int_to_32_bytes(x):
        return x.to_bytes(x.bit_length(), byteorder="little")[0:32]

    a = int_to_32_bytes(x) + int_to_32_bytes(y)

    keccak_hash = keccak.new(digest_bits=256)
    keccak_hash.update(a)
    r = keccak_hash.hexdigest()
    return r[12 * 2 :]


def generate_certificates(private_key, public_key):
    today = datetime.datetime.today()
    one_day = datetime.timedelta(1, 0, 0)
    address = public_address(public_key)  # .map(Base16.encode).getOrElse("local")
    owner = f"CN={address}"

    builder = x509.CertificateBuilder()
    builder = builder.not_valid_before(today)

    # TODO: Where's documentation of the decision to make keys valid for 1 year only?
    builder = builder.not_valid_after(today + 365 * one_day)
    builder = builder.subject_name(
        x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, owner)])
    )
    builder = builder.issuer_name(
        x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, owner)])
    )
    builder = builder.public_key(public_key)
    builder = builder.serial_number(x509.random_serial_number())
    certificate = builder.sign(
        private_key=private_key, algorithm=hashes.SHA256(), backend=default_backend()
    )

    cert_pem = certificate.public_bytes(encoding=serialization.Encoding.PEM)
    key_pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )
    return cert_pem, key_pem


def blake2b_hash(data: bytes) -> bytes:
    h = blake2b(digest_size=32)
    h.update(data)
    return h.digest()


def signature(private_key, data: bytes):
    return private_key and consensus.Signature(
        sig_algorithm="ed25519",
        sig=ed25519.SigningKey(read_pem_key(private_key)).sign(data),
    )


def private_to_public_key(private_key) -> bytes:
    return ed25519.SigningKey(read_pem_key(private_key)).get_verifying_key().to_bytes()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: casper.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import annotations_pb2 as annotations__pb2
from . import empty_pb2 as empty__pb2
from . import consensus_pb2 as consensus__pb2
from . import info_pb2 as info__pb2
from . import state_pb2 as state__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='casper.proto',
  package='io.casperlabs.node.api.casper',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=b'\n\x0c\x63\x61sper.proto\x12\x1dio.casperlabs.node.api.casper\x1a\x11\x61nnotations.proto\x1a\x0b\x65mpty.proto\x1a\x0f\x63onsensus.proto\x1a\ninfo.proto\x1a\x0bstate.proto\"G\n\rDeployRequest\x12\x36\n\x06\x64\x65ploy\x18\x01 \x01(\x0b\x32&.io.casperlabs.casper.consensus.Deploy\"\x87\x01\n\x13GetBlockInfoRequest\x12\x19\n\x11\x62lock_hash_base16\x18\x01 \x01(\t\x12\x12\n\nblock_hash\x18\x03 \x01(\x0c\x12\x41\n\x04view\x18\x02 \x01(\x0e\x32\x33.io.casperlabs.casper.consensus.info.BlockInfo.View\"}\n\x17StreamBlockInfosRequest\x12\r\n\x05\x64\x65pth\x18\x01 \x01(\r\x12\x10\n\x08max_rank\x18\x02 \x01(\x04\x12\x41\n\x04view\x18\x03 \x01(\x0e\x32\x33.io.casperlabs.casper.consensus.info.BlockInfo.View\"\x8b\x01\n\x14GetDeployInfoRequest\x12\x1a\n\x12\x64\x65ploy_hash_base16\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65ploy_hash\x18\x03 \x01(\x0c\x12\x42\n\x04view\x18\x02 \x01(\x0e\x32\x34.io.casperlabs.casper.consensus.info.DeployInfo.View\"\x8e\x01\n\x19StreamBlockDeploysRequest\x12\x19\n\x11\x62lock_hash_base16\x18\x01 \x01(\t\x12\x12\n\nblock_hash\x18\x03 \x01(\x0c\x12\x42\n\x04view\x18\x02 \x01(\x0e\x32\x34.io.casperlabs.casper.consensus.info.DeployInfo.View\"\xd9\x01\n\nStateQuery\x12I\n\x0bkey_variant\x18\x01 \x01(\x0e\x32\x34.io.casperlabs.node.api.casper.StateQuery.KeyVariant\x12\x12\n\nkey_base16\x18\x02 \x01(\t\x12\x15\n\rpath_segments\x18\x03 \x03(\t\"U\n\nKeyVariant\x12\x1b\n\x17KEY_VARIANT_UNSPECIFIED\x10\x00\x12\x08\n\x04HASH\x10\x01\x12\x08\n\x04UREF\x10\x02\x12\x0b\n\x07\x41\x44\x44RESS\x10\x03\x12\t\n\x05LOCAL\x10\x04\"\x7f\n\x14GetBlockStateRequest\x12\x19\n\x11\x62lock_hash_base16\x18\x01 \x01(\t\x12\x12\n\nblock_hash\x18\x03 \x01(\x0c\x12\x38\n\x05query\x18\x02 \x01(\x0b\x32).io.casperlabs.node.api.casper.StateQuery\"\x86\x01\n\x19\x42\x61tchGetBlockStateRequest\x12\x19\n\x11\x62lock_hash_base16\x18\x01 \x01(\t\x12\x12\n\nblock_hash\x18\x03 \x01(\x0c\x12:\n\x07queries\x18\x02 \x03(\x0b\x32).io.casperlabs.node.api.casper.StateQuery\"m\n\x1a\x42\x61tchGetBlockStateResponse\x12I\n\x06values\x18\x02 \x03(\x0b\x32\x39.io.casperlabs.casper.consensus.state.StoredValueInstanceJ\x04\x08\x01\x10\x02\"\xc2\x01\n\x16ListDeployInfosRequest\x12!\n\x19\x61\x63\x63ount_public_key_base16\x18\x01 \x01(\t\x12\x1a\n\x12\x61\x63\x63ount_public_key\x18\x05 \x01(\x0c\x12\x42\n\x04view\x18\x02 \x01(\x0e\x32\x34.io.casperlabs.casper.consensus.info.DeployInfo.View\x12\x11\n\tpage_size\x18\x03 \x01(\r\x12\x12\n\npage_token\x18\x04 \x01(\t\"\x92\x01\n\x17ListDeployInfosResponse\x12\x45\n\x0c\x64\x65ploy_infos\x18\x01 \x03(\x0b\x32/.io.casperlabs.casper.consensus.info.DeployInfo\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\x12\x17\n\x0fprev_page_token\x18\x03 \x01(\t\"e\n GetLastFinalizedBlockInfoRequest\x12\x41\n\x04view\x18\x01 \x01(\x0e\x32\x33.io.casperlabs.casper.consensus.info.BlockInfo.View\"\x8b\x03\n\x13StreamEventsRequest\x12\x13\n\x0b\x62lock_added\x18\x01 \x01(\x08\x12\x17\n\x0f\x62lock_finalized\x18\x02 \x01(\x08\x12\x14\n\x0c\x64\x65ploy_added\x18\x03 \x01(\x08\x12\x18\n\x10\x64\x65ploy_discarded\x18\x04 \x01(\x08\x12\x17\n\x0f\x64\x65ploy_requeued\x18\x05 \x01(\x08\x12\x18\n\x10\x64\x65ploy_processed\x18\x06 \x01(\x08\x12\x18\n\x10\x64\x65ploy_finalized\x18\x07 \x01(\x08\x12\x17\n\x0f\x64\x65ploy_orphaned\x18\x08 \x01(\x08\x12V\n\rdeploy_filter\x18\t \x01(\x0b\x32?.io.casperlabs.node.api.casper.StreamEventsRequest.DeployFilter\x12\x14\n\x0cmin_event_id\x18\n \x01(\x04\x1a\x42\n\x0c\x44\x65ployFilter\x12\x1b\n\x13\x61\x63\x63ount_public_keys\x18\x01 \x03(\x0c\x12\x15\n\rdeploy_hashes\x18\x02 \x03(\x0c\x32\xcd\x0b\n\rCasperService\x12g\n\x06\x44\x65ploy\x12,.io.casperlabs.node.api.casper.DeployRequest\x1a\x16.google.protobuf.Empty\"\x17\x82\xd3\xe4\x93\x02\x11\"\x07\x64\x65ploys:\x06\x64\x65ploy\x12\x91\x01\n\x0cGetBlockInfo\x12\x32.io.casperlabs.node.api.casper.GetBlockInfoRequest\x1a..io.casperlabs.casper.consensus.info.BlockInfo\"\x1d\x82\xd3\xe4\x93\x02\x17\x12\x15{block_hash_base16=*}\x12\x8c\x01\n\x10StreamBlockInfos\x12\x36.io.casperlabs.node.api.casper.StreamBlockInfosRequest\x1a..io.casperlabs.casper.consensus.info.BlockInfo\"\x0e\x82\xd3\xe4\x93\x02\x08\x12\x06\x62locks0\x01\x12\x95\x01\n\rGetDeployInfo\x12\x33.io.casperlabs.node.api.casper.GetDeployInfoRequest\x1a/.io.casperlabs.casper.consensus.info.DeployInfo\"\x1e\x82\xd3\xe4\x93\x02\x18\x12\x16{deploy_hash_base16=*}\x12\x98\x01\n\x12StreamBlockDeploys\x12\x38.io.casperlabs.node.api.casper.StreamBlockDeploysRequest\x1a\x35.io.casperlabs.casper.consensus.Block.ProcessedDeploy\"\x0f\x82\xd3\xe4\x93\x02\t\x12\x07\x64\x65ploys0\x01\x12\x80\x01\n\x0cStreamEvents\x12\x32.io.casperlabs.node.api.casper.StreamEventsRequest\x1a*.io.casperlabs.casper.consensus.info.Event\"\x0e\x82\xd3\xe4\x93\x02\x08\x12\x06\x65vents0\x01\x12\x91\x01\n\rGetBlockState\x12\x33.io.casperlabs.node.api.casper.GetBlockStateRequest\x1a\x39.io.casperlabs.casper.consensus.state.StoredValueInstance\"\x10\x82\xd3\xe4\x93\x02\n\"\x05state:\x01*\x12\xa4\x01\n\x12\x42\x61tchGetBlockState\x12\x38.io.casperlabs.node.api.casper.BatchGetBlockStateRequest\x1a\x39.io.casperlabs.node.api.casper.BatchGetBlockStateResponse\"\x19\x82\xd3\xe4\x93\x02\x13\"\x0estate:batchGet:\x01*\x12\x91\x01\n\x0fListDeployInfos\x12\x35.io.casperlabs.node.api.casper.ListDeployInfosRequest\x1a\x36.io.casperlabs.node.api.casper.ListDeployInfosResponse\"\x0f\x82\xd3\xe4\x93\x02\t\x12\x07\x64\x65ploys\x12\xaa\x01\n\x19GetLastFinalizedBlockInfo\x12?.io.casperlabs.node.api.casper.GetLastFinalizedBlockInfoRequest\x1a..io.casperlabs.casper.consensus.info.BlockInfo\"\x1c\x82\xd3\xe4\x93\x02\x16\x12\x14last-finalized-blockb\x06proto3'
  ,
  dependencies=[annotations__pb2.DESCRIPTOR,empty__pb2.DESCRIPTOR,consensus__pb2.DESCRIPTOR,info__pb2.DESCRIPTOR,state__pb2.DESCRIPTOR,])



_STATEQUERY_KEYVARIANT = _descriptor.EnumDescriptor(
  name='KeyVariant',
  full_name='io.casperlabs.node.api.casper.StateQuery.KeyVariant',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='KEY_VARIANT_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='HASH', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UREF', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ADDRESS', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LOCAL', index=4, number=4,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=879,
  serialized_end=964,
)
_sym_db.RegisterEnumDescriptor(_STATEQUERY_KEYVARIANT)


_DEPLOYREQUEST = _descriptor.Descriptor(
  name='DeployRequest',
  full_name='io.casperlabs.node.api.casper.DeployRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='deploy', full_name='io.casperlabs.node.api.casper.DeployRequest.deploy', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=121,
  serialized_end=192,
)


_GETBLOCKINFOREQUEST = _descriptor.Descriptor(
  name='GetBlockInfoRequest',
  full_name='io.casperlabs.node.api.casper.GetBlockInfoRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='block_hash_base16', full_name='io.casperlabs.node.api.casper.GetBlockInfoRequest.block_hash_base16', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_hash', full_name='io.casperlabs.node.api.casper.GetBlockInfoRequest.block_hash', index=1,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='view', full_name='io.casperlabs.node.api.casper.GetBlockInfoRequest.view', index=2,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=195,
  serialized_end=330,
)


_STREAMBLOCKINFOSREQUEST = _descriptor.Descriptor(
  name='StreamBlockInfosRequest',
  full_name='io.casperlabs.node.api.casper.StreamBlockInfosRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='depth', full_name='io.casperlabs.node.api.casper.StreamBlockInfosRequest.depth', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_rank', full_name='io.casperlabs.node.api.casper.StreamBlockInfosRequest.max_rank', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='view', full_name='io.casperlabs.node.api.casper.StreamBlockInfosRequest.view', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=332,
  serialized_end=457,
)


_GETDEPLOYINFOREQUEST = _descriptor.Descriptor(
  name='GetDeployInfoRequest',
  full_name='io.casperlabs.node.api.casper.GetDeployInfoRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='deploy_hash_base16', full_name='io.casperlabs.node.api.casper.GetDeployInfoRequest.deploy_hash_base16', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='deploy_hash', full_name='io.casperlabs.node.api.casper.GetDeployInfoRequest.deploy_hash', index=1,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='view', full_name='io.casperlabs.node.api.casper.GetDeployInfoRequest.view', index=2,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=460,
  serialized_end=599,
)


_STREAMBLOCKDEPLOYSREQUEST = _descriptor.Descriptor(
  name='StreamBlockDeploysRequest',
  full_name='io.casperlabs.node.api.casper.StreamBlockDeploysRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='block_hash_base16', full_name='io.casperlabs.node.api.casper.StreamBlockDeploysRequest.block_hash_base16', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_hash', full_name='io.casperlabs.node.api.casper.StreamBlockDeploysRequest.block_hash', index=1,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='view', full_name='io.casperlabs.node.api.casper.StreamBlockDeploysRequest.view', index=2,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=602,
  serialized_end=744,
)


_STATEQUERY = _descriptor.Descriptor(
  name='StateQuery',
  full_name='io.casperlabs.node.api.casper.StateQuery',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key_variant', full_name='io.casperlabs.node.api.casper.StateQuery.key_variant', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='key_base16', full_name='io.casperlabs.node.api.casper.StateQuery.key_base16', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='path_segments', full_name='io.casperlabs.node.api.casper.StateQuery.path_segments', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _STATEQUERY_KEYVARIANT,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=747,
  serialized_end=964,
)


_GETBLOCKSTATEREQUEST = _descriptor.Descriptor(
  name='GetBlockStateRequest',
  full_name='io.casperlabs.node.api.casper.GetBlockStateRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='block_hash_base16', full_name='io.casperlabs.node.api.casper.GetBlockStateRequest.block_hash_base16', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_hash', full_name='io.casperlabs.node.api.casper.GetBlockStateRequest.block_hash', index=1,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='query', full_name='io.casperlabs.node.api.casper.GetBlockStateRequest.query', index=2,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=966,
  serialized_end=1093,
)


_BATCHGETBLOCKSTATEREQUEST = _descriptor.Descriptor(
  name='BatchGetBlockStateRequest',
  full_name='io.casperlabs.node.api.casper.BatchGetBlockStateRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='block_hash_base16', full_name='io.casperlabs.node.api.casper.BatchGetBlockStateRequest.block_hash_base16', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_hash', full_name='io.casperlabs.node.api.casper.BatchGetBlockStateRequest.block_hash', index=1,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='queries', full_name='io.casperlabs.node.api.casper.BatchGetBlockStateRequest.queries', index=2,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1096,
  serialized_end=1230,
)


_BATCHGETBLOCKSTATERESPONSE = _descriptor.Descriptor(
  name='BatchGetBlockStateResponse',
  full_name='io.casperlabs.node.api.casper.BatchGetBlockStateResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='values', full_name='io.casperlabs.node.api.casper.BatchGetBlockStateResponse.values', index=0,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1232,
  serialized_end=1341,
)


_LISTDEPLOYINFOSREQUEST = _descriptor.Descriptor(
  name='ListDeployInfosRequest',
  full_name='io.casperlabs.node.api.casper.ListDeployInfosRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='account_public_key_base16', full_name='io.casperlabs.node.api.casper.ListDeployInfosRequest.account_public_key_base16', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='account_public_key', full_name='io.casperlabs.node.api.casper.ListDeployInfosRequest.account_public_key', index=1,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='view', full_name='io.casperlabs.node.api.casper.ListDeployInfosRequest.view', index=2,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='io.casperlabs.node.api.casper.ListDeployInfosRequest.page_size', index=3,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='io.casperlabs.node.api.casper.ListDeployInfosRequest.page_token', index=4,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1344,
  serialized_end=1538,
)


_LISTDEPLOYINFOSRESPONSE = _descriptor.Descriptor(
  name='ListDeployInfosResponse',
  full_name='io.casperlabs.node.api.casper.ListDeployInfosResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='deploy_infos', full_name='io.casperlabs.node.api.casper.ListDeployInfosResponse.deploy_infos', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='io.casperlabs.node.api.casper.ListDeployInfosResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='prev_page_token', full_name='io.casperlabs.node.api.casper.ListDeployInfosResponse.prev_page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1541,
  serialized_end=1687,
)


_GETLASTFINALIZEDBLOCKINFOREQUEST = _descriptor.Descriptor(
  name='GetLastFinalizedBlockInfoRequest',
  full_name='io.casperlabs.node.api.casper.GetLastFinalizedBlockInfoRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='view', full_name='io.casperlabs.node.api.casper.GetLastFinalizedBlockInfoRequest.view', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1689,
  serialized_end=1790,
)


_STREAMEVENTSREQUEST_DEPLOYFILTER = _descriptor.Descriptor(
  name='DeployFilter',
  full_name='io.casperlabs.node.api.casper.StreamEventsRequest.DeployFilter',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='account_public_keys', full_name='io.casperlabs.node.api.casper.StreamEventsRequest.DeployFilter.account_public_keys', index=0,
      number=1, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='deploy_hashes', full_name='io.casperlabs.node.api.casper.StreamEventsRequest.DeployFilter.deploy_hashes', index=1,
      number=2, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2122,
  serialized_end=2188,
)

_STREAMEVENTSREQUEST = _descriptor.Descriptor(
  name='StreamEventsRequest',
  full_name='io.casperlabs.node.api.casper.StreamEventsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='block_added', full_name='io.casperlabs.node.api.casper.StreamEventsRequest.block_added', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_finalized', full_name='io.casperlabs.node.api.casper.StreamEventsRequest.block_finalized', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='deploy_added', full_name='io.casperlabs.node.api.casper.StreamEventsRequest.deploy_added', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='deploy_discarded', full_name='io.casperlabs.node.api.casper.StreamEventsRequest.deploy_discarded', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='deploy_requeued', full_name='io.casperlabs.node.api.casper.StreamEventsRequest.deploy_requeued', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='deploy_processed', full_name='io.casperlabs.node.api.casper.StreamEventsRequest.deploy_processed', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='deploy_finalized', full_name='io.casperlabs.node.api.casper.StreamEventsRequest.deploy_finalized', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='deploy_orphaned', full_name='io.casperlabs.node.api.casper.StreamEventsRequest.deploy_orphaned', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='deploy_filter', full_name='io.casperlabs.node.api.casper.StreamEventsRequest.deploy_filter', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='min_event_id', full_name='io.casperlabs.node.api.casper.StreamEventsRequest.min_event_id', index=9,
      number=10, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_STREAMEVENTSREQUEST_DEPLOYFILTER, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1793,
  serialized_end=2188,
)

_DEPLOYREQUEST.fields_by_name['deploy'].message_type = consensus__pb2._DEPLOY
_GETBLOCKINFOREQUEST.fields_by_name['view'].enum_type = info__pb2._BLOCKINFO_VIEW
_STREAMBLOCKINFOSREQUEST.fields_by_name['view'].enum_type = info__pb2._BLOCKINFO_VIEW
_GETDEPLOYINFOREQUEST.fields_by_name['view'].enum_type = info__pb2._DEPLOYINFO_VIEW
_STREAMBLOCKDEPLOYSREQUEST.fields_by_name['view'].enum_type = info__pb2._DEPLOYINFO_VIEW
_STATEQUERY.fields_by_name['key_variant'].enum_type = _STATEQUERY_KEYVARIANT
_STATEQUERY_KEYVARIANT.containing_type = _STATEQUERY
_GETBLOCKSTATEREQUEST.fields_by_name['query'].message_type = _STATEQUERY
_BATCHGETBLOCKSTATEREQUEST.fields_by_name['queries'].message_type = _STATEQUERY
_BATCHGETBLOCKSTATERESPONSE.fields_by_name['values'].message_type = state__pb2._STOREDVALUEINSTANCE
_LISTDEPLOYINFOSREQUEST.fields_by_name['view'].enum_type = info__pb2._DEPLOYINFO_VIEW
_LISTDEPLOYINFOSRESPONSE.fields_by_name['deploy_infos'].message_type = info__pb2._DEPLOYINFO
_GETLASTFINALIZEDBLOCKINFOREQUEST.fields_by_name['view'].enum_type = info__pb2._BLOCKINFO_VIEW
_STREAMEVENTSREQUEST_DEPLOYFILTER.containing_type = _STREAMEVENTSREQUEST
_STREAMEVENTSREQUEST.fields_by_name['deploy_filter'].message_type = _STREAMEVENTSREQUEST_DEPLOYFILTER
DESCRIPTOR.message_types_by_name['DeployRequest'] = _DEPLOYREQUEST
DESCRIPTOR.message_types_by_name['GetBlockInfoRequest'] = _GETBLOCKINFOREQUEST
DESCRIPTOR.message_types_by_name['StreamBlockInfosRequest'] = _STREAMBLOCKINFOSREQUEST
DESCRIPTOR.message_types_by_name['GetDeployInfoRequest'] = _GETDEPLOYINFOREQUEST
DESCRIPTOR.message_types_by_name['StreamBlockDeploysRequest'] = _STREAMBLOCKDEPLOYSREQUEST
DESCRIPTOR.message_types_by_name['StateQuery'] = _STATEQUERY
DESCRIPTOR.message_types_by_name['GetBlockStateRequest'] = _GETBLOCKSTATEREQUEST
DESCRIPTOR.message_types_by_name['BatchGetBlockStateRequest'] = _BATCHGETBLOCKSTATEREQUEST
DESCRIPTOR.message_types_by_name['BatchGetBlockStateResponse'] = _BATCHGETBLOCKSTATERESPONSE
DESCRIPTOR.message_types_by_name['ListDeployInfosRequest'] = _LISTDEPLOYINFOSREQUEST
DESCRIPTOR.message_types_by_name['ListDeployInfosResponse'] = _LISTDEPLOYINFOSRESPONSE
DESCRIPTOR.message_types_by_name['GetLastFinalizedBlockInfoRequest'] = _GETLASTFINALIZEDBLOCKINFOREQUEST
DESCRIPTOR.message_types_by_name['StreamEventsRequest'] = _STREAMEVENTSREQUEST
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DeployRequest = _reflection.GeneratedProtocolMessageType('DeployRequest', (_message.Message,), {
  'DESCRIPTOR' : _DEPLOYREQUEST,
  '__module__' : 'casper_pb2'
  # @@protoc_insertion_point(class_scope:io.casperlabs.node.api.casper.DeployRequest)
  })
_sym_db.RegisterMessage(DeployRequest)

GetBlockInfoRequest = _reflection.GeneratedProtocolMessageType('GetBlockInfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETBLOCKINFOREQUEST,
  '__module__' : 'casper_pb2'
  # @@protoc_insertion_point(class_scope:io.casperlabs.node.api.casper.GetBlockInfoRequest)
  })
_sym_db.RegisterMessage(GetBlockInfoRequest)

StreamBlockInfosRequest = _reflection.GeneratedProtocolMessageType('StreamBlockInfosRequest', (_message.Message,), {
  'DESCRIPTOR' : _STREAMBLOCKINFOSREQUEST,
  '__module__' : 'casper_pb2'
  # @@protoc_insertion_point(class_scope:io.casperlabs.node.api.casper.StreamBlockInfosRequest)
  })
_sym_db.RegisterMessage(StreamBlockInfosRequest)

GetDeployInfoRequest = _reflection.GeneratedProtocolMessageType('GetDeployInfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETDEPLOYINFOREQUEST,
  '__module__' : 'casper_pb2'
  # @@protoc_insertion_point(class_scope:io.casperlabs.node.api.casper.GetDeployInfoRequest)
  })
_sym_db.RegisterMessage(GetDeployInfoRequest)

StreamBlockDeploysRequest = _reflection.GeneratedProtocolMessageType('StreamBlockDeploysRequest', (_message.Message,), {
  'DESCRIPTOR' : _STREAMBLOCKDEPLOYSREQUEST,
  '__module__' : 'casper_pb2'
  # @@protoc_insertion_point(class_scope:io.casperlabs.node.api.casper.StreamBlockDeploysRequest)
  })
_sym_db.RegisterMessage(StreamBlockDeploysRequest)

StateQuery = _reflection.GeneratedProtocolMessageType('StateQuery', (_message.Message,), {
  'DESCRIPTOR' : _STATEQUERY,
  '__module__' : 'casper_pb2'
  # @@protoc_insertion_point(class_scope:io.casperlabs.node.api.casper.StateQuery)
  })
_sym_db.RegisterMessage(StateQuery)

GetBlockStateRequest = _reflection.GeneratedProtocolMessageType('GetBlockStateRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETBLOCKSTATEREQUEST,
  '__module__' : 'casper_pb2'
  # @@protoc_insertion_point(class_scope:io.casperlabs.node.api.casper.GetBlockStateRequest)
  })
_sym_db.RegisterMessage(GetBlockStateRequest)

BatchGetBlockStateRequest = _reflection.GeneratedProtocolMessageType('BatchGetBlockStateRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETBLOCKSTATEREQUEST,
  '__module__' : 'casper_pb2'
  # @@protoc_insertion_point(class_scope:io.casperlabs.node.api.casper.BatchGetBlockStateRequest)
  })
_sym_db.RegisterMessage(BatchGetBlockStateRequest)

BatchGetBlockStateResponse = _reflection.GeneratedProtocolMessageType('BatchGetBlockStateResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETBLOCKSTATERESPONSE,
  '__module__' : 'casper_pb2'
  # @@protoc_insertion_point(class_scope:io.casperlabs.node.api.casper.BatchGetBlockStateResponse)
  })
_sym_db.RegisterMessage(BatchGetBlockStateResponse)

ListDeployInfosRequest = _reflection.GeneratedProtocolMessageType('ListDeployInfosRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTDEPLOYINFOSREQUEST,
  '__module__' : 'casper_pb2'
  # @@protoc_insertion_point(class_scope:io.casperlabs.node.api.casper.ListDeployInfosRequest)
  })
_sym_db.RegisterMessage(ListDeployInfosRequest)

ListDeployInfosResponse = _reflection.GeneratedProtocolMessageType('ListDeployInfosResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTDEPLOYINFOSRESPONSE,
  '__module__' : 'casper_pb2'
  # @@protoc_insertion_point(class_scope:io.casperlabs.node.api.casper.ListDeployInfosResponse)
  })
_sym_db.RegisterMessage(ListDeployInfosResponse)

GetLastFinalizedBlockInfoRequest = _reflection.GeneratedProtocolMessageType('GetLastFinalizedBlockInfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETLASTFINALIZEDBLOCKINFOREQUEST,
  '__module__' : 'casper_pb2'
  # @@protoc_insertion_point(class_scope:io.casperlabs.node.api.casper.GetLastFinalizedBlockInfoRequest)
  })
_sym_db.RegisterMessage(GetLastFinalizedBlockInfoRequest)

StreamEventsRequest = _reflection.GeneratedProtocolMessageType('StreamEventsRequest', (_message.Message,), {

  'DeployFilter' : _reflection.GeneratedProtocolMessageType('DeployFilter', (_message.Message,), {
    'DESCRIPTOR' : _STREAMEVENTSREQUEST_DEPLOYFILTER,
    '__module__' : 'casper_pb2'
    # @@protoc_insertion_point(class_scope:io.casperlabs.node.api.casper.StreamEventsRequest.DeployFilter)
    })
  ,
  'DESCRIPTOR' : _STREAMEVENTSREQUEST,
  '__module__' : 'casper_pb2'
  # @@protoc_insertion_point(class_scope:io.casperlabs.node.api.casper.StreamEventsRequest)
  })
_sym_db.RegisterMessage(StreamEventsRequest)
_sym_db.RegisterMessage(StreamEventsRequest.DeployFilter)



_CASPERSERVICE = _descriptor.ServiceDescriptor(
  name='CasperService',
  full_name='io.casperlabs.node.api.casper.CasperService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=2191,
  serialized_end=3676,
  methods=[
  _descriptor.MethodDescriptor(
    name='Deploy',
    full_name='io.casperlabs.node.api.casper.CasperService.Deploy',
    index=0,
    containing_service=None,
    input_type=_DEPLOYREQUEST,
    output_type=empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\021\"\007deploys:\006deploy',
  ),
  _descriptor.MethodDescriptor(
    name='GetBlockInfo',
    full_name='io.casperlabs.node.api.casper.CasperService.GetBlockInfo',
    index=1,
    containing_service=None,
    input_type=_GETBLOCKINFOREQUEST,
    output_type=info__pb2._BLOCKINFO,
    serialized_options=b'\202\323\344\223\002\027\022\025{block_hash_base16=*}',
  ),
  _descriptor.MethodDescriptor(
    name='StreamBlockInfos',
    full_name='io.casperlabs.node.api.casper.CasperService.StreamBlockInfos',
    index=2,
    containing_service=None,
    input_type=_STREAMBLOCKINFOSREQUEST,
    output_type=info__pb2._BLOCKINFO,
    serialized_options=b'\202\323\344\223\002\010\022\006blocks',
  ),
  _descriptor.MethodDescriptor(
    name='GetDeployInfo',
    full_name='io.casperlabs.node.api.casper.CasperService.GetDeployInfo',
    index=3,
    containing_service=None,
    input_type=_GETDEPLOYINFOREQUEST,
    output_type=info__pb2._DEPLOYINFO,
    serialized_options=b'\202\323\344\223\002\030\022\026{deploy_hash_base16=*}',
  ),
  _descriptor.MethodDescriptor(
    name='StreamBlockDeploys',
    full_name='io.casperlabs.node.api.casper.CasperService.StreamBlockDeploys',
    index=4,
    containing_service=None,
    input_type=_STREAMBLOCKDEPLOYSREQUEST,
    output_type=consensus__pb2._BLOCK_PROCESSEDDEPLOY,
    serialized_options=b'\202\323\344\223\002\t\022\007deploys',
  ),
  _descriptor.MethodDescriptor(
    name='StreamEvents',
    full_name='io.casperlabs.node.api.casper.CasperService.StreamEvents',
    index=5,
    containing_service=None,
    input_type=_STREAMEVENTSREQUEST,
    output_type=info__pb2._EVENT,
    serialized_options=b'\202\323\344\223\002\010\022\006events',
  ),
  _descriptor.MethodDescriptor(
    name='GetBlockState',
    full_name='io.casperlabs.node.api.casper.CasperService.GetBlockState',
    index=6,
    containing_service=None,
    input_type=_GETBLOCKSTATEREQUEST,
    output_type=state__pb2._STOREDVALUEINSTANCE,
    serialized_options=b'\202\323\344\223\002\n\"\005state:\001*',
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetBlockState',
    full_name='io.casperlabs.node.api.casper.CasperService.BatchGetBlockState',
    index=7,
    containing_service=None,
    input_type=_BATCHGETBLOCKSTATEREQUEST,
    output_type=_BATCHGETBLOCKSTATERESPONSE,
    serialized_options=b'\202\323\344\223\002\023\"\016state:batchGet:\001*',
  ),
  _descriptor.MethodDescriptor(
    name='ListDeployInfos',
    full_name='io.casperlabs.node.api.casper.CasperService.ListDeployInfos',
    index=8,
    containing_service=None,
    input_type=_LISTDEPLOYINFOSREQUEST,
    output_type=_LISTDEPLOYINFOSRESPONSE,
    serialized_options=b'\202\323\344\223\002\t\022\007deploys',
  ),
  _descriptor.MethodDescriptor(
    name='GetLastFinalizedBlockInfo',
    full_name='io.casperlabs.node.api.casper.CasperService.GetLastFinalizedBlockInfo',
    index=9,
    containing_service=None,
    input_type=_GETLASTFINALIZEDBLOCKINFOREQUEST,
    output_type=info__pb2._BLOCKINFO,
    serialized_options=b'\202\323\344\223\002\026\022\024last-finalized-block',
  ),
])
_sym_db.RegisterServiceDescriptor(_CASPERSERVICE)

DESCRIPTOR.services_by_name['CasperService'] = _CASPERSERVICE

# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python3
"""
CasperLabs Client API library and command line tool.
"""

from pathlib import Path

# Hack to fix the relative imports problems #
import sys

sys.path.append(str(Path(__file__).resolve().parents[1]))
# end of hack #
import os
import time
import grpc
from grpc._channel import _Rendezvous
import functools
import logging
import tempfile

# Monkey patching of google.protobuf.text_encoding.CEscape
# to get keys and signatures in hex when printed
import google.protobuf.text_format

CEscape = google.protobuf.text_format.text_encoding.CEscape


def _hex(text, as_utf8):
    try:
        return (len(text) in (32, 64, 20)) and text.hex() or CEscape(text, as_utf8)
    except TypeError:
        return CEscape(text, as_utf8)


google.protobuf.text_format.text_encoding.CEscape = _hex

# ~/CasperLabs/protobuf/io/casperlabs/node/api/control.proto
from . import control_pb2_grpc
from . import control_pb2 as control

# ~/CasperLabs/protobuf/io/casperlabs/node/api/casper.proto
from . import casper_pb2 as casper
from . import casper_pb2_grpc

# ~/CasperLabs/protobuf/io/casperlabs/casper/consensus/info.proto
from . import info_pb2 as info

# ~/CasperLabs/protobuf/io/casperlabs/node/api/diagnostics.proto
from . import diagnostics_pb2_grpc
from . import empty_pb2

from . import vdag
from . import abi
from casperlabs_client.utils import (
    bundled_contract,
    extract_common_name,
    key_variant,
    make_deploy,
    sign_deploy,
    get_public_key,
)


DEFAULT_HOST = "localhost"
DEFAULT_PORT = 40401
DEFAULT_INTERNAL_PORT = 40402


class InternalError(Exception):
    """
    The only exception that API calls can throw.
    Internal errors like gRPC exceptions will be caught
    and this exception thrown instead, so the user does
    not have to worry about handling any other exceptions.
    """

    def __init__(self, status="", details=""):
        super(InternalError, self).__init__()
        self.status = status
        self.details = details

    def __str__(self):
        return f"{self.status}: {self.details}"


def api(function):
    """
    Decorator of API functions that protects user code from
    unknown exceptions raised by gRPC or internal API errors.
    It will catch all exceptions and throw InternalError.

    :param function: function to be decorated
    :return:
    """

    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except (SyntaxError, TypeError, InternalError):
            raise
        except _Rendezvous as e:
            raise InternalError(str(e.code()), e.details())
        except Exception as e:
            raise InternalError(details=str(e)) from e

    return wrapper


NUMBER_OF_RETRIES = 5

# Initial delay in seconds before an attempt to retry
INITIAL_DELAY = 0.3


def retry_wrapper(function, *args):
    delay = INITIAL_DELAY
    for i in range(NUMBER_OF_RETRIES):
        try:
            return function(*args)
        except _Rendezvous as e:
            if e.code() == grpc.StatusCode.UNAVAILABLE and i < NUMBER_OF_RETRIES - 1:
                delay += delay
                logging.warning(f"Retrying after {e} in {delay} seconds")
                time.sleep(delay)
            else:
                raise


def retry_unary(function):
    @functools.wraps(function)
    def wrapper(*args):
        return retry_wrapper(function, *args)

    return wrapper


def retry_stream(function):
    @functools.wraps(function)
    def wrapper(*args):
        yield from retry_wrapper(function, *args)

    return wrapper


class InsecureGRPCService:
    def __init__(self, host, port, serviceStub):
        self.address = f"{host}:{port}"
        self.serviceStub = serviceStub

    def __getattr__(self, name):

        logging.debug(
            f"Creating insecure connection to {self.address} ({self.serviceStub})"
        )

        @retry_unary
        def unary_unary(*args):
            logging.debug(
                f"Insecure {self.address} ({self.serviceStub}): {name} {list(args)}"
            )
            with grpc.insecure_channel(self.address) as channel:
                return getattr(self.serviceStub(channel), name)(*args)

        @retry_stream
        def unary_stream(*args):
            logging.debug(
                f"Insecure {self.address} ({self.serviceStub}): {name} {list(args)}"
            )
            with grpc.insecure_channel(self.address) as channel:
                yield from getattr(self.serviceStub(channel), name[: -len("_stream")])(
                    *args
                )

        return name.endswith("_stream") and unary_stream or unary_unary


class SecureGRPCService:
    def __init__(self, host, port, serviceStub, node_id, certificate_file):
        self.address = f"{host}:{port}"
        self.serviceStub = serviceStub
        self.node_id = node_id or extract_common_name(certificate_file)
        self.certificate_file = certificate_file
        with open(self.certificate_file, "rb") as f:
            self.credentials = grpc.ssl_channel_credentials(f.read())
        self.secure_channel_options = self.node_id and (
            ("grpc.ssl_target_name_override", self.node_id),
            ("grpc.default_authority", self.node_id),
        )

    def __getattr__(self, name):
        logging.debug(
            f"Creating secure connection to {self.address} ({self.serviceStub})"
        )

        @retry_unary
        def unary_unary(*args):
            with grpc.secure_channel(
                self.address, self.credentials, options=self.secure_channel_options
            ) as channel:
                return getattr(self.serviceStub(channel), name)(*args)

        @retry_stream
        def unary_stream(*args):
            with grpc.secure_channel(
                self.address, self.credentials, options=self.secure_channel_options
            ) as channel:
                yield from getattr(self.serviceStub(channel), name[: -len("_stream")])(
                    *args
                )

        return name.endswith("_stream") and unary_stream or unary_unary


class CasperLabsClient:
    """
    gRPC CasperLabs client.
    """

    DEPLOY_STATUS_CHECK_DELAY = 0.5
    DEPLOY_STATUS_TIMEOUT = 180  # 3 minutes

    def __init__(
        self,
        host: str = DEFAULT_HOST,
        port: int = DEFAULT_PORT,
        port_internal: int = DEFAULT_INTERNAL_PORT,
        node_id: str = None,
        certificate_file: str = None,
    ):
        """
        CasperLabs client's constructor.

        :param host:            Hostname or IP of node on which gRPC service is running
        :param port:            Port used for external gRPC API
        :param port_internal:   Port used for internal gRPC API
        :param certificate_file:      Certificate file for TLS
        :param node_id:         node_id of the node, for gRPC encryption
        """
        self.host = host
        self.port = port
        self.port_internal = port_internal
        self.node_id = node_id
        self.certificate_file = certificate_file

        if node_id:
            self.casperService = SecureGRPCService(
                host, port, casper_pb2_grpc.CasperServiceStub, node_id, certificate_file
            )
            self.controlService = SecureGRPCService(
                # We currently assume that if node_id is given then
                # we get certificate_file too. This is unlike in the Scala client
                # where node_id is all that's needed for configuring secure connection.
                # The reason for this is that currently it doesn't seem to be possible
                # to open a secure grpc connection in Python without supplying any
                # certificate on the client side.
                host,
                port_internal,
                control_pb2_grpc.ControlServiceStub,
                node_id,
                certificate_file,
            )
            self.diagnosticsService = SecureGRPCService(
                host,
                port,
                diagnostics_pb2_grpc.DiagnosticsStub,
                node_id,
                certificate_file,
            )

        else:
            self.casperService = InsecureGRPCService(
                host, port, casper_pb2_grpc.CasperServiceStub
            )
            self.controlService = InsecureGRPCService(
                host, port_internal, control_pb2_grpc.ControlServiceStub
            )
            self.diagnosticsService = InsecureGRPCService(
                host, port, diagnostics_pb2_grpc.DiagnosticsStub
            )

    @api
    def make_deploy(
        self,
        from_addr: bytes = None,
        gas_price: int = 10,
        payment: str = None,
        session: str = None,
        public_key: str = None,
        session_args: bytes = None,
        payment_args: bytes = None,
        payment_amount: int = None,
        payment_hash: bytes = None,
        payment_name: str = None,
        payment_uref: bytes = None,
        session_hash: bytes = None,
        session_name: str = None,
        session_uref: bytes = None,
        ttl_millis: int = 0,
        dependencies: list = None,
        chain_name: str = None,
    ):
        """
        Create a protobuf deploy object. See deploy for description of parameters.
        """
        return make_deploy(
            from_addr=from_addr,
            gas_price=gas_price,
            payment=payment,
            session=session,
            public_key=public_key,
            session_args=session_args,
            payment_args=payment_args,
            payment_amount=payment_amount,
            payment_hash=payment_hash,
            payment_name=payment_name,
            payment_uref=payment_uref,
            session_hash=session_hash,
            session_name=session_name,
            session_uref=session_uref,
            ttl_millis=ttl_millis,
            dependencies=dependencies,
            chain_name=chain_name,
        )

    @api
    def sign_deploy(self, deploy, public_key, private_key_file):
        return sign_deploy(deploy, public_key, private_key_file)

    @api
    def deploy(
        self,
        from_addr: bytes = None,
        gas_price: int = 10,
        payment: str = None,
        session: str = None,
        public_key: str = None,
        private_key: str = None,
        session_args: bytes = None,
        payment_args: bytes = None,
        payment_amount: int = None,
        payment_hash: bytes = None,
        payment_name: str = None,
        payment_uref: bytes = None,
        session_hash: bytes = None,
        session_name: str = None,
        session_uref: bytes = None,
        ttl_millis: int = 0,
        dependencies=None,
        chain_name: str = None,
    ):
        """
        Deploy a smart contract source file to Casper on an existing running node.
        The deploy will be packaged and sent as a block to the network depending
        on the configuration of the Casper instance.

        :param from_addr:     Purse address that will be used to pay for the deployment.
        :param gas_price:     The price of gas for this transaction in units dust/gas.
                              Must be positive integer.
        :param payment:       Path to the file with payment code.
        :param session:       Path to the file with session code.
        :param public_key:    Path to a file with public key (Ed25519)
        :param private_key:   Path to a file with private key (Ed25519)
        :param session_args:  List of ABI encoded arguments of session contract
        :param payment_args:  List of ABI encoded arguments of payment contract
        :param session-hash:  Hash of the stored contract to be called in the
                              session; base16 encoded.
        :param session-name:  Name of the stored contract (associated with the
                              executing account) to be called in the session.
        :param session-uref:  URef of the stored contract to be called in the
                              session; base16 encoded.
        :param payment-hash:  Hash of the stored contract to be called in the
                              payment; base16 encoded.
        :param payment-name:  Name of the stored contract (associated with the
                              executing account) to be called in the payment.
        :param payment-uref:  URef of the stored contract to be called in the
                              payment; base16 encoded.
        :ttl_millis:          Time to live. Time (in milliseconds) that the
                              deploy will remain valid for.
        :dependencies:        List of deploy hashes (base16 encoded) which
                              must be executed before this deploy.
        :chain_name:          Name of the chain to optionally restrict the
                              deploy from being accidentally included
                              anywhere else.
        :return:              deploy hash in base16 format
        """

        deploy = self.make_deploy(
            from_addr=from_addr,
            gas_price=gas_price,
            payment=payment,
            session=session,
            public_key=public_key,
            session_args=session_args,
            payment_args=payment_args,
            payment_amount=payment_amount,
            payment_hash=payment_hash,
            payment_name=payment_name,
            payment_uref=payment_uref,
            session_hash=session_hash,
            session_name=session_name,
            session_uref=session_uref,
            ttl_millis=ttl_millis,
            dependencies=dependencies,
            chain_name=chain_name,
        )

        deploy = self.sign_deploy(
            deploy, get_public_key(public_key, from_addr, private_key), private_key
        )
        self.send_deploy(deploy)
        return deploy.deploy_hash.hex()

    @api
    def transfer(self, target_account_hex, amount, **deploy_args):
        target_account_bytes = bytes.fromhex(target_account_hex)
        deploy_args["session"] = bundled_contract("transfer_to_account_u512.wasm")
        deploy_args["session_args"] = abi.ABI.args(
            [
                abi.ABI.account("account", target_account_bytes),
                abi.ABI.u512("amount", amount),
            ]
        )
        return self.deploy(**deploy_args)

    @api
    def send_deploy(self, deploy):
        self.casperService.Deploy(casper.DeployRequest(deploy=deploy))

    @api
    def showBlocks(self, depth: int = 1, max_rank=0, full_view=True):
        """
        Get slices of the DAG, going backwards, rank by rank.

        :param depth:     How many of the top ranks of the DAG to show.
        :param max_rank:  Maximum rank to go back from.
                          0 means go from the current tip of the DAG.
        :param full_view: Full view if True, otherwise basic.
        :return:          Generator of block info objects.
        """
        yield from self.casperService.StreamBlockInfos_stream(
            casper.StreamBlockInfosRequest(
                depth=depth,
                max_rank=max_rank,
                view=(
                    full_view and info.BlockInfo.View.FULL or info.BlockInfo.View.BASIC
                ),
            )
        )

    @api
    def showBlock(self, block_hash_base16: str, full_view=True):
        """
        Returns object describing a block known by Casper on an existing running node.

        :param block_hash_base16: hash of the block to be retrieved
        :param full_view:         full view if True, otherwise basic
        :return:                  object representing the retrieved block
        """
        return self.casperService.GetBlockInfo(
            casper.GetBlockInfoRequest(
                block_hash_base16=block_hash_base16,
                view=(
                    full_view and info.BlockInfo.View.FULL or info.BlockInfo.View.BASIC
                ),
            )
        )

    @api
    def propose(self):
        """"
        THIS METHOD IS DEPRECATED! It will be removed soon.

        Propose a block using deploys in the pool.

        :return:    response object with block_hash
        """
        return self.controlService.Propose(control.ProposeRequest())

    @api
    def visualizeDag(
        self,
        depth: int,
        out: str = None,
        show_justification_lines: bool = False,
        stream: str = None,
        delay_in_seconds=5,
    ):
        """
        Generate DAG in DOT format.

        :param depth:                     depth in terms of block height
        :param out:                       output image filename, outputs to stdout if
                                          not specified, must end with one of the png,
                                          svg, svg_standalone, xdot, plain, plain_ext,
                                          ps, ps2, json, json0
        :param show_justification_lines:  if justification lines should be shown
        :param stream:                    subscribe to changes, 'out' has to specified,
                                          valid values are 'single-output', 'multiple-outputs'
        :param delay_in_seconds:          delay in seconds when polling for updates (streaming)
        :return:                          Yields generated DOT source or file name when out provided.
                                          Generates endless stream of file names if stream is not None.
        """
        block_infos = list(self.showBlocks(depth, full_view=False))
        dot_dag_description = vdag.generate_dot(block_infos, show_justification_lines)
        if not out:
            yield dot_dag_description
            return

        parts = out.split(".")
        file_format = parts[-1]
        file_name_base = ".".join(parts[:-1])

        iteration = -1

        def file_name():
            nonlocal iteration, file_format
            if not stream or stream == "single-output":
                return out
            else:
                iteration += 1
                return f"{file_name_base}_{iteration}.{file_format}"

        yield self._call_dot(dot_dag_description, file_name(), file_format)
        previous_block_hashes = set(b.summary.block_hash for b in block_infos)
        while stream:
            time.sleep(delay_in_seconds)
            block_infos = list(self.showBlocks(depth, full_view=False))
            block_hashes = set(b.summary.block_hash for b in block_infos)
            if block_hashes != previous_block_hashes:
                dot_dag_description = vdag.generate_dot(
                    block_infos, show_justification_lines
                )
                yield self._call_dot(dot_dag_description, file_name(), file_format)
                previous_block_hashes = block_hashes

    def _call_dot(self, dot_dag_description, file_name, file_format):
        with tempfile.NamedTemporaryFile(mode="w") as f:
            f.write(dot_dag_description)
            f.flush()
            cmd = f"dot -T{file_format} -o {file_name} {f.name}"
            rc = os.system(cmd)
            if rc:
                raise Exception(f"Call to dot ({cmd}) failed with error code {rc}")
        print(f"Wrote {file_name}")
        return file_name

    @api
    def queryState(self, blockHash: str, key: str, path: str, keyType: str):
        """
        Query a value in the global state.

        :param blockHash:         Hash of the block to query the state of
        :param key:               Base16 encoding of the base key
        :param path:              Path to the value to query. Must be of the form
                                  'key1/key2/.../keyn'
        :param keyType:           Type of base key. Must be one of 'hash', 'uref', 'address' or 'local'.
                                  For 'local' key type, 'key' value format is {seed}:{rest},
                                  where both parts are hex encoded."
        :return:                  QueryStateResponse object
        """
        q = casper.StateQuery(key_variant=key_variant(keyType), key_base16=key)
        q.path_segments.extend([name for name in path.split("/") if name])
        return self.casperService.GetBlockState(
            casper.GetBlockStateRequest(block_hash_base16=blockHash, query=q)
        )

    @api
    def balance(self, address: str, block_hash: str):
        value = self.queryState(block_hash, address, "", "address")
        account = None
        try:
            account = value.account
        except AttributeError:
            return InternalError(
                "balance", f"Expected Account type value under {address}."
            )

        urefs = [u for u in account.named_keys if u.name == "mint"]
        if len(urefs) == 0:
            raise InternalError(
                "balance",
                "Account's named_keys map did not contain Mint contract address.",
            )

        mintPublic = urefs[0]

        mintPublicHex = mintPublic.key.uref.uref.hex()
        purseAddrHex = account.main_purse.uref.hex()
        localKeyValue = f"{mintPublicHex}:{purseAddrHex}"

        balanceURef = self.queryState(block_hash, localKeyValue, "", "local")
        balanceURefHex = balanceURef.cl_value.value.key.uref.uref.hex()
        balance = self.queryState(block_hash, balanceURefHex, "", "uref")
        balanceStrValue = balance.cl_value.value.u512.value
        return int(balanceStrValue)

    @api
    def showDeploy(
        self,
        deploy_hash_base16: str,
        full_view: bool = False,
        wait_for_processed: bool = False,
        delay: int = DEPLOY_STATUS_CHECK_DELAY,
        timeout_seconds: int = DEPLOY_STATUS_TIMEOUT,
    ):
        """
        Retrieve information about a single deploy by hash.
        """
        start_time = time.time()
        while True:
            deploy_info = self.casperService.GetDeployInfo(
                casper.GetDeployInfoRequest(
                    deploy_hash_base16=deploy_hash_base16,
                    view=(
                        full_view
                        and info.DeployInfo.View.FULL
                        or info.DeployInfo.View.BASIC
                    ),
                )
            )
            if (
                wait_for_processed
                and deploy_info.status.state == info.DeployInfo.State.PENDING
            ):
                if time.time() - start_time > timeout_seconds:
                    raise Exception(
                        f"Timed out waiting for deploy {deploy_hash_base16} to be processed"
                    )
                time.sleep(delay)
                continue
            return deploy_info

    @api
    def showDeploys(self, block_hash_base16: str, full_view=True):
        """
        Get the processed deploys within a block.
        """
        yield from self.casperService.StreamBlockDeploys_stream(
            casper.StreamBlockDeploysRequest(
                block_hash_base16=block_hash_base16,
                view=(
                    full_view
                    and info.DeployInfo.View.FULL
                    or info.DeployInfo.View.BASIC
                ),
            )
        )

    @api
    def stream_events(
        self,
        all: bool = False,
        block_added: bool = False,
        block_finalized: bool = False,
        deploy_added: bool = False,
        deploy_discarded: bool = False,
        deploy_requeued: bool = False,
        deploy_processed: bool = False,
        deploy_finalized: bool = False,
        deploy_orphaned: bool = False,
        account_public_keys=None,
        deploy_hashes=None,
        min_event_id: int = 0,
    ):
        """
        See StreamEventsRequest in
            ~/CasperLabs/protobuf/io/casperlabs/node/api/casper.proto
        for description of types of events.

        Note, you must subscribe to some events (pass True to some keywords other than account_public_keys or deploy_hashes)
        otherwise this generator will block forever.
        """
        if all:
            block_added = True
            block_finalized = True
            deploy_added = True
            deploy_discarded = True
            deploy_requeued = True
            deploy_processed = True
            deploy_finalized = True
            deploy_orphaned = True

        yield from self.casperService.StreamEvents_stream(
            casper.StreamEventsRequest(
                block_added=block_added,
                block_finalized=block_finalized,
                deploy_added=deploy_added,
                deploy_discarded=deploy_discarded,
                deploy_requeued=deploy_requeued,
                deploy_processed=deploy_processed,
                deploy_finalized=deploy_finalized,
                deploy_orphaned=deploy_orphaned,
                deploy_filter=casper.StreamEventsRequest.DeployFilter(
                    account_public_keys=[
                        bytes.fromhex(pk) for pk in account_public_keys or []
                    ],
                    deploy_hashes=[bytes.fromhex(h) for h in deploy_hashes or []],
                ),
                min_event_id=min_event_id,
            )
        )

    @api
    def wait_for_deploy_processed(
        self,
        deploy_hash,
        on_error_raise=True,
        delay=DEPLOY_STATUS_CHECK_DELAY,
        timeout_seconds=DEPLOY_STATUS_TIMEOUT,
    ):
        start_time = time.time()
        result = None
        while True:
            if time.time() - start_time > timeout_seconds:
                raise Exception(
                    f"Timed out waiting for deploy {deploy_hash} to be processed"
                )

            result = self.showDeploy(deploy_hash, full_view=False)
            if result.status.state != info.DeployInfo.State.PENDING:
                break
            time.sleep(delay)

        if on_error_raise:
            if len(result.processing_results) == 0:
                raise Exception(f"Deploy {deploy_hash} status: {result.status}")

            last_processing_result = result.processing_results[0]
            if last_processing_result.is_error:
                raise Exception(
                    f"Deploy {deploy_hash} execution error: {last_processing_result.error_message}"
                )
        return result

    @api
    def show_peers(self):
        return list(self.diagnosticsService.ListPeers(empty_pb2.Empty()).peers)

    def cli(self, *arguments) -> int:
        from . import cli

        return cli(
            "--host",
            self.host,
            "--port",
            self.port,
            "--port-internal",
            self.port_internal,
            *arguments,
        )



"""
Command line interface for CasperLabsClient.
"""

import argparse
import textwrap
import base64
import sys
import os
import functools
import logging
from pathlib import Path
import datetime
from casperlabs_client import (
    CasperLabsClient,
    DEFAULT_HOST,
    DEFAULT_PORT,
    DEFAULT_INTERNAL_PORT,
    bundled_contract,
)
from casperlabs_client.utils import hexify
from casperlabs_client.abi import ABI
from casperlabs_client.crypto import (
    read_pem_key,
    generate_validators_keys,
    generate_key_pair,
    public_address,
    private_to_public_key,
    generate_certificates,
)
from . import consensus_pb2 as consensus

DEFAULT_PAYMENT_AMOUNT = 10000000

DOT_FORMATS = "canon,cmap,cmapx,cmapx_np,dot,dot_json,eps,fig,gd,gd2,gif,gv,imap,imap_np,ismap,jpe,jpeg,jpg,json,json0,mp,pdf,pic,plain,plain-ext,png,pov,ps,ps2,svg,svgz,tk,vml,vmlz,vrml,wbmp,x11,xdot,xdot1.2,xdot1.4,xdot_json,xlib"


def guarded_command(function):
    """
    Decorator of functions that implement CLI commands.

    Occasionally the node can throw some exceptions instead of properly sending us a response,
    those will be deserialized on our end and rethrown by the gRPC layer.
    In this case we want to catch the exception and return a non-zero return code to the shell.

    :param function:  function to be decorated
    :return:
    """

    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        try:
            rc = function(*args, **kwargs)
            # Generally the CLI commands are assumed to succeed if they don't throw,
            # but they can also return a positive error code if they need to.
            if rc is not None:
                return rc
            return 0
        except Exception as e:
            print(str(e), file=sys.stderr)
            return 1

    return wrapper


def _show_blocks(response, element_name="block"):
    count = 0
    for block in response:
        print(f"------------- {element_name} {count} ---------------")
        print(hexify(block))
        print("-----------------------------------------------------\n")
        count += 1
    print("count:", count)


def _show_block(response):
    print(hexify(response))


def _set_session(args, file_name):
    """
    Use bundled contract unless one of the session* args is set.
    """
    if not any((args.session, args.session_hash, args.session_name, args.session_uref)):
        args.session = bundled_contract(file_name)


@guarded_command
def bond_command(casperlabs_client, args):
    logging.info(f"BOND {args}")
    _set_session(args, "bonding.wasm")

    if not args.session_args:
        args.session_args = ABI.args_to_json(
            ABI.args([ABI.long_value("amount", args.amount)])
        )

    return deploy_command(casperlabs_client, args)


@guarded_command
def unbond_command(casperlabs_client, args):
    logging.info(f"UNBOND {args}")
    _set_session(args, "unbonding.wasm")

    if not args.session_args:
        args.session_args = ABI.args_to_json(
            ABI.args(
                [ABI.optional_value("amount", ABI.long_value("amount", args.amount))]
            )
        )

    return deploy_command(casperlabs_client, args)


@guarded_command
def transfer_command(casperlabs_client, args):
    _set_session(args, "transfer_to_account_u512.wasm")

    if not args.session_args:
        target_account_bytes = base64.b64decode(args.target_account)
        if len(target_account_bytes) != 32:
            target_account_bytes = bytes.fromhex(args.target_account)
            if len(target_account_bytes) != 32:
                raise Exception(
                    "--target_account must be 32 bytes base64 or base16 encoded"
                )

        args.session_args = ABI.args_to_json(
            ABI.args(
                [
                    ABI.account("account", target_account_bytes),
                    ABI.u512("amount", args.amount),
                ]
            )
        )

    return deploy_command(casperlabs_client, args)


def _deploy_kwargs(args, private_key_accepted=True):
    from_addr = (
        getattr(args, "from")
        and bytes.fromhex(getattr(args, "from"))
        or getattr(args, "public_key")
        and read_pem_key(args.public_key)
        or private_to_public_key(args.private_key)
    )
    if from_addr and len(from_addr) != 32:
        raise Exception(
            "--from must be 32 bytes encoded as 64 characters long hexadecimal"
        )

    if not (args.payment_amount or args.payment_args):
        args.payment_amount = DEFAULT_PAYMENT_AMOUNT

    if args.payment_amount:
        args.payment_args = ABI.args_to_json(
            ABI.args([ABI.big_int("amount", int(args.payment_amount))])
        )
        # Unless one of payment* options supplied use bundled standard-payment
        if not any(
            (args.payment, args.payment_name, args.payment_hash, args.payment_uref)
        ):
            args.payment = bundled_contract("standard_payment.wasm")

    d = dict(
        from_addr=from_addr,
        gas_price=args.gas_price,
        payment=args.payment or args.session,
        session=args.session,
        public_key=args.public_key or None,
        session_args=args.session_args
        and ABI.args_from_json(args.session_args)
        or None,
        payment_args=args.payment_args
        and ABI.args_from_json(args.payment_args)
        or None,
        payment_hash=args.payment_hash and bytes.fromhex(args.payment_hash),
        payment_name=args.payment_name,
        payment_uref=args.payment_uref and bytes.fromhex(args.payment_uref),
        session_hash=args.session_hash and bytes.fromhex(args.session_hash),
        session_name=args.session_name,
        session_uref=args.session_uref and bytes.fromhex(args.session_uref),
        ttl_millis=args.ttl_millis,
        dependencies=args.dependencies,
        chain_name=args.chain_name,
    )
    if private_key_accepted:
        d["private_key"] = args.private_key or None
    return d


@guarded_command
def make_deploy_command(casperlabs_client, args):
    kwargs = _deploy_kwargs(args, private_key_accepted=False)
    deploy = casperlabs_client.make_deploy(**kwargs)
    data = deploy.SerializeToString()
    if not args.deploy_path:
        sys.stdout.buffer.write(data)
    else:
        with open(args.deploy_path, "wb") as f:
            f.write(data)


@guarded_command
def sign_deploy_command(casperlabs_client, args):
    deploy = consensus.Deploy()
    if args.deploy_path:
        with open(args.deploy_path, "rb") as input_file:
            deploy.ParseFromString(input_file.read())
    else:
        deploy.ParseFromString(sys.stdin.read())

    deploy = casperlabs_client.sign_deploy(
        deploy, read_pem_key(args.public_key), args.private_key
    )

    if not args.signed_deploy_path:
        sys.stdout.write(deploy.SerializeToString())
    else:
        with open(args.signed_deploy_path, "wb") as output_file:
            output_file.write(deploy.SerializeToString())


@guarded_command
def send_deploy_command(casperlabs_client, args):
    deploy = consensus.Deploy()
    with open(args.deploy_path, "rb") as f:
        deploy.ParseFromString(f.read())
        casperlabs_client.send_deploy(deploy)
    print(f"Success! Deploy {deploy.deploy_hash.hex()} deployed")


@guarded_command
def deploy_command(casperlabs_client, args):
    kwargs = _deploy_kwargs(args)
    deploy_hash = casperlabs_client.deploy(**kwargs)
    print(f"Success! Deploy {deploy_hash} deployed")
    if args.wait_for_processed:
        deploy_info = casperlabs_client.showDeploy(
            deploy_hash,
            full_view=False,
            wait_for_processed=args.wait_for_processed,
            timeout_seconds=args.timeout_seconds,
        )
        print(hexify(deploy_info))


@guarded_command
def propose_command(casperlabs_client, args):
    print("Warning: command propose is deprecated.", file=sys.stderr)
    response = casperlabs_client.propose()
    print(f"Success! Block hash: {response.block_hash.hex()}")


@guarded_command
def show_block_command(casperlabs_client, args):
    response = casperlabs_client.showBlock(args.hash, full_view=True)
    return _show_block(response)


@guarded_command
def show_blocks_command(casperlabs_client, args):
    response = casperlabs_client.showBlocks(args.depth, full_view=False)
    _show_blocks(response)


@guarded_command
def vdag_command(casperlabs_client, args):
    for o in casperlabs_client.visualizeDag(
        args.depth, args.out, args.show_justification_lines, args.stream
    ):
        if not args.out:
            print(o)
            break


@guarded_command
def query_state_command(casperlabs_client, args):

    response = casperlabs_client.queryState(
        args.block_hash, args.key, args.path or "", getattr(args, "type")
    )
    print(hexify(response))


@guarded_command
def balance_command(casperlabs_client, args):
    response = casperlabs_client.balance(args.address, args.block_hash)
    print(response)


@guarded_command
def show_deploy_command(casperlabs_client, args):
    response = casperlabs_client.showDeploy(
        args.hash,
        full_view=False,
        wait_for_processed=args.wait_for_processed,
        timeout_seconds=args.timeout_seconds,
    )
    print(hexify(response))


@guarded_command
def show_deploys_command(casperlabs_client, args):
    response = casperlabs_client.showDeploys(args.hash, full_view=False)
    _show_blocks(response, element_name="deploy")


def write_file(file_name, text):
    with open(file_name, "w") as f:
        f.write(text)


def write_file_binary(file_name, data):
    with open(file_name, "wb") as f:
        f.write(data)


def encode_base64(a: bytes):
    return str(base64.b64encode(a), "utf-8")


@guarded_command
def keygen_command(casperlabs_client, args):
    directory = Path(args.directory).resolve()
    validator_private_path = directory / "validator-private.pem"
    validator_pub_path = directory / "validator-public.pem"
    validator_id_path = directory / "validator-id"
    validator_id_hex_path = directory / "validator-id-hex"
    node_priv_path = directory / "node.key.pem"
    node_cert_path = directory / "node.certificate.pem"
    node_id_path = directory / "node-id"

    validator_private_pem, validator_public_pem, validator_public_bytes = (
        generate_validators_keys()
    )
    write_file_binary(validator_private_path, validator_private_pem)
    write_file_binary(validator_pub_path, validator_public_pem)
    write_file(validator_id_path, encode_base64(validator_public_bytes))
    write_file(validator_id_hex_path, validator_public_bytes.hex())

    private_key, public_key = generate_key_pair()

    node_cert, key_pem = generate_certificates(private_key, public_key)

    write_file_binary(node_priv_path, key_pem)
    write_file_binary(node_cert_path, node_cert)

    write_file(node_id_path, public_address(public_key))
    print(f"Keys successfully created in directory: {str(directory.absolute())}")


@guarded_command
def show_peers_command(casperlabs_client, args):
    peers = casperlabs_client.show_peers()
    i = 0
    for i, node in enumerate(peers, 1):
        print(f"------------- node {i} ---------------")
        print(hexify(node))
    print("-----------------------------------------------------")
    print(f"count: {i}")


@guarded_command
def stream_events_command(casperlabs_client, args):
    subscribed_events = dict(
        all=args.all,
        block_added=args.block_added,
        block_finalized=args.block_finalized,
        deploy_added=args.deploy_added,
        deploy_discarded=args.deploy_discarded,
        deploy_requeued=args.deploy_requeued,
        deploy_processed=args.deploy_processed,
        deploy_finalized=args.deploy_finalized,
        deploy_orphaned=args.deploy_orphaned,
    )
    if not any(subscribed_events.values()):
        raise argparse.ArgumentTypeError("No events chosen")

    stream = casperlabs_client.stream_events(
        account_public_keys=args.account_public_key,
        deploy_hashes=args.deploy_hash,
        min_event_id=args.min_event_id,
        **subscribed_events,
    )
    for event in stream:
        now = datetime.datetime.now()
        print(f"------------- {now.strftime('%Y-%m-%d %H:%M:%S')} -------------")
        print(hexify(event))


def check_directory(path):
    if not os.path.exists(path):
        raise argparse.ArgumentTypeError(f"Directory '{path}' does not exist")
    if not os.path.isdir(path):
        raise argparse.ArgumentTypeError(f"'{path}' is not a directory")
    if not os.access(path, os.W_OK):
        raise argparse.ArgumentTypeError(f"'{path}' does not have writing permissions")
    return Path(path)


def dot_output(file_name):
    """
    Check file name has an extension of one of file formats supported by Graphviz.
    """
    parts = file_name.split(".")
    if len(parts) == 1:
        raise argparse.ArgumentTypeError(
            f"'{file_name}' has no extension indicating file format"
        )
    else:
        file_format = parts[-1]
        if file_format not in DOT_FORMATS.split(","):
            raise argparse.ArgumentTypeError(
                f"File extension {file_format} not recognized, must be one of {DOT_FORMATS}"
            )
    return file_name


def natural(number):
    """Check number is an integer greater than 0"""
    n = int(number)
    if n < 1:
        raise argparse.ArgumentTypeError(f"{number} is not a positive int value")
    return n


# fmt: off
def deploy_options(private_key_accepted=True):
    return ([
        [('-f', '--from'), dict(required=False, type=str, help="The public key of the account which is the context of this deployment, base16 encoded.")],
        [('--chain-name',), dict(required=False, type=str, help="Name of the chain to optionally restrict the deploy from being accidentally included anywhere else.")],
        [('--dependencies',), dict(required=False, nargs="+", default=None, help="List of deploy hashes (base16 encoded) which must be executed before this deploy.")],
        [('--payment-amount',), dict(required=False, type=int, default=None, help="Standard payment amount. Use this with the default payment, or override with --payment-args if custom payment code is used. By default --payment-amount is set to 10000000")],
        [('--gas-price',), dict(required=False, type=int, default=10, help='The price of gas for this transaction in units dust/gas. Must be positive integer.')],
        [('-p', '--payment'), dict(required=False, type=str, default=None, help='Path to the file with payment code, by default fallbacks to the --session code')],
        [('--payment-hash',), dict(required=False, type=str, default=None, help='Hash of the stored contract to be called in the payment; base16 encoded')],
        [('--payment-name',), dict(required=False, type=str, default=None, help='Name of the stored contract (associated with the executing account) to be called in the payment')],
        [('--payment-uref',), dict(required=False, type=str, default=None, help='URef of the stored contract to be called in the payment; base16 encoded')],
        [('-s', '--session'), dict(required=False, type=str, default=None, help='Path to the file with session code')],
        [('--session-hash',), dict(required=False, type=str, default=None, help='Hash of the stored contract to be called in the session; base16 encoded')],
        [('--session-name',), dict(required=False, type=str, default=None, help='Name of the stored contract (associated with the executing account) to be called in the session')],
        [('--session-uref',), dict(required=False, type=str, default=None, help='URef of the stored contract to be called in the session; base16 encoded')],
        [('--session-args',), dict(required=False, type=str, help="""JSON encoded list of session args, e.g.: '[{"name": "amount", "value": {"long_value": 123456}}]'""")],
        [('--payment-args',), dict(required=False, type=str, help="""JSON encoded list of payment args, e.g.: '[{"name": "amount", "value": {"big_int": {"value": "123456", "bit_width": 512}}}]'""")],
        [('--ttl-millis',), dict(required=False, type=int, help="""Time to live. Time (in milliseconds) that the deploy will remain valid for.'""")],
        [('-w', '--wait-for-processed'), dict(action='store_true', help='Wait for deploy status PROCESSED or DISCARDED')],
        [('--timeout-seconds',), dict(type=int, default=CasperLabsClient.DEPLOY_STATUS_TIMEOUT, help='Timeout in seconds')],
        [('--public-key',), dict(required=False, default=None, type=str, help='Path to the file with account public key (Ed25519)')]]
        + (private_key_accepted
           and [[('--private-key',), dict(required=True, default=None, type=str, help='Path to the file with account private key (Ed25519)')]]
           or []))
# fmt:on


def cli(*arguments) -> int:
    """
    Parse list of command line arguments and call appropriate command.
    """

    class Parser:
        def __init__(self):
            # The --help option added by default has a short version -h, which conflicts
            # with short version of --host, so we need to disable it.
            self.parser = argparse.ArgumentParser(
                prog="casperlabs_client", add_help=False
            )
            self.parser.add_argument(
                "--help",
                action="help",
                default=argparse.SUPPRESS,
                help="show this help message and exit",
            )
            self.parser.add_argument(
                "-h",
                "--host",
                required=False,
                default=DEFAULT_HOST,
                type=str,
                help="Hostname or IP of node on which gRPC service is running.",
            )
            self.parser.add_argument(
                "-p",
                "--port",
                required=False,
                default=DEFAULT_PORT,
                type=int,
                help="Port used for external gRPC API.",
            )
            self.parser.add_argument(
                "--port-internal",
                required=False,
                default=DEFAULT_INTERNAL_PORT,
                type=int,
                help="Port used for internal gRPC API.",
            )
            self.parser.add_argument(
                "--node-id",
                required=False,
                type=str,
                help="node_id parameter for TLS connection",
            )
            self.parser.add_argument(
                "--certificate-file",
                required=False,
                type=str,
                help="Certificate file for TLS connection",
            )
            self.sp = self.parser.add_subparsers(help="Choose a request")

            def no_command(casperlabs_client, args):
                print(
                    "You must provide a command. --help for documentation of commands."
                )
                self.parser.print_usage()
                return 1

            self.parser.set_defaults(function=no_command)

        def addCommand(self, command: str, function, help, arguments):
            command_parser = self.sp.add_parser(command, help=help)
            command_parser.set_defaults(function=function)
            for (args, options) in arguments:
                command_parser.add_argument(*args, **options)

        def run(self, argv):
            args = self.parser.parse_args(argv)
            return args.function(
                CasperLabsClient(
                    args.host,
                    args.port,
                    args.port_internal,
                    args.node_id,
                    args.certificate_file,
                ),
                args,
            )

    parser = Parser()

    # fmt: off
    parser.addCommand('deploy', deploy_command, 'Deploy a smart contract source file to Casper on an existing running node. The deploy will be packaged and sent as a block to the network depending on the configuration of the Casper instance',
                      deploy_options())

    parser.addCommand('make-deploy', make_deploy_command, "Constructs a deploy that can be signed and sent to a node.",
                      [[('-o', '--deploy-path'), dict(required=False, help="Path to the file where deploy will be saved. Optional, if not provided the deploy will be printed to STDOUT.")]] + deploy_options(private_key_accepted=False))

    parser.addCommand('sign-deploy', sign_deploy_command, "Cryptographically signs a deploy. The signature is appended to existing approvals.",
                      [[('-o', '--signed-deploy-path'), dict(required=False, default=None, help="Path to the file where signed deploy will be saved. Optional, if not provided the deploy will be printed to STDOUT.")],
                       [('-i', '--deploy-path'), dict(required=False, default=None, help="Path to the deploy file.")],
                       [('--private-key',), dict(required=True, help="Path to the file with account private key (Ed25519)")],
                       [('--public-key',), dict(required=True, help="Path to the file with account public key (Ed25519)")]])

    parser.addCommand('send-deploy', send_deploy_command, "Deploy a smart contract source file to Casper on an existing running node. The deploy will be packaged and sent as a block to the network depending on the configuration of the Casper instance.",
                      [[('-i', '--deploy-path'), dict(required=False, default=None, help="Path to the file with signed deploy.")]])

    parser.addCommand('bond', bond_command, 'Issues bonding request',
                      [[('-a', '--amount'), dict(required=True, type=int, help='amount of motes to bond')]] + deploy_options())

    parser.addCommand('unbond', unbond_command, 'Issues unbonding request',
                      [[('-a', '--amount'),
                       dict(required=False, default=None, type=int, help='Amount of motes to unbond. If not provided then a request to unbond with full staked amount is made.')]] + deploy_options())

    parser.addCommand('transfer', transfer_command, 'Transfers funds between accounts',
                      [[('-a', '--amount'), dict(required=True, default=None, type=int, help='Amount of motes to transfer. Note: a mote is the smallest, indivisible unit of a token.')],
                       [('-t', '--target-account'), dict(required=True, type=str, help="base64 or base16 representation of target account's public key")],
                       ] + deploy_options(private_key_accepted=True))

    parser.addCommand('propose', propose_command, '[DEPRECATED] Force a node to propose a block based on its accumulated deploys.', [])

    parser.addCommand('show-block', show_block_command, 'View properties of a block known by Casper on an existing running node. Output includes: parent hashes, storage contents of the tuplespace.',
                      [[('hash',), dict(type=str, help='the hash value of the block')]])

    parser.addCommand('show-blocks', show_blocks_command, 'View list of blocks in the current Casper view on an existing running node.',
                      [[('-d', '--depth'), dict(required=True, type=int, help='depth in terms of block height')]])

    parser.addCommand('show-deploy', show_deploy_command, 'View properties of a deploy known by Casper on an existing running node.',
                      [[('hash',), dict(type=str, help='Value of the deploy hash, base16 encoded.')],
                       [('-w', '--wait-for-processed'), dict(action='store_true', help='Wait for deploy status PROCESSED or DISCARDED')],
                       [('--timeout-seconds',), dict(type=int, default=CasperLabsClient.DEPLOY_STATUS_TIMEOUT, help='Timeout in seconds')]])

    parser.addCommand('show-deploys', show_deploys_command, 'View deploys included in a block.',
                      [[('hash',), dict(type=str, help='Value of the block hash, base16 encoded.')]])

    parser.addCommand('vdag', vdag_command, 'DAG in DOT format. You need to install Graphviz from https://www.graphviz.org/ to use it.',
                      [[('-d', '--depth'), dict(required=True, type=natural, help='depth in terms of block height')],
                       [('-o', '--out'), dict(required=False, type=dot_output, help=f'output image filename, outputs to stdout if not specified, must end with one of {DOT_FORMATS}')],
                       [('-s', '--show-justification-lines'), dict(action='store_true', help='if justification lines should be shown')],
                       [('--stream',), dict(required=False, choices=('single-output', 'multiple-outputs'), help="subscribe to changes, '--out' has to be specified, valid values are 'single-output', 'multiple-outputs'")]])

    parser.addCommand('query-state', query_state_command, 'Query a value in the global state.',
                      [[('-b', '--block-hash'), dict(required=True, type=str, help='Hash of the block to query the state of')],
                       [('-k', '--key'), dict(required=True, type=str, help='Base16 encoding of the base key')],
                       [('-p', '--path'), dict(required=False, type=str, help="Path to the value to query. Must be of the form 'key1/key2/.../keyn'")],
                       [('-t', '--type'), dict(required=True, choices=('hash', 'uref', 'address', 'local'), help="Type of base key. Must be one of 'hash', 'uref', 'address' or 'local'. For 'local' key type, 'key' value format is {seed}:{rest}, where both parts are hex encoded.")]])

    parser.addCommand('balance', balance_command, 'Returns the balance of the account at the specified block.',
                      [[('-a', '--address'), dict(required=True, type=str, help="Account's public key in hex.")],
                       [('-b', '--block-hash'), dict(required=True, type=str, help='Hash of the block to query the state of')]])

    parser.addCommand('keygen', keygen_command, textwrap.dedent("""\
         Generate keys.

         Usage: casperlabs-client keygen <existingOutputDirectory>
         Command will override existing files!
         Generated files:
           node-id               # node ID as in casperlabs://c0a6c82062461c9b7f9f5c3120f44589393edf31@<NODE ADDRESS>?protocol=40400&discovery=40404
                                 # derived from node.key.pem
           node.certificate.pem  # TLS certificate used for node-to-node interaction encryption
                                 # derived from node.key.pem
           node.key.pem          # secp256r1 private key
           validator-id          # validator ID in Base64 format; can be used in accounts.csv
                                 # derived from validator.public.pem
           validator-id-hex      # validator ID in hex, derived from validator.public.pem
           validator-private.pem # ed25519 private key
           validator-public.pem  # ed25519 public key"""),
                      [[('directory',), dict(type=check_directory, help="Output directory for keys. Should already exists.")]])

    parser.addCommand('show-peers', show_peers_command, "Show peers connected to the node.", [])

    parser.addCommand('stream-events', stream_events_command, "Stream block and deploy state transition events.", [
        [('--all',), dict(action='store_true', help='Subscribe to all events')],
        [('--block-added',), dict(action='store_true', help='Block added')],
        [('--block-finalized',), dict(action='store_true', help='Block finalized')],
        [('--deploy-added',), dict(action='store_true', help='Deploy added')],
        [('--deploy-discarded',), dict(action='store_true', help='Deploy discarded')],
        [('--deploy-requeued',), dict(action='store_true', help='Deploy requeued')],
        [('--deploy-processed',), dict(action='store_true', help='Deploy processed')],
        [('--deploy-finalized',), dict(action='store_true', help='Deploy finalized')],
        [('--deploy-orphaned',), dict(action='store_true', help='Deploy orphaned')],
        [('-k', '--account-public-key'), dict(action='append', help='Filter by (possibly multiple) account public key(s)')],
        [('-d', '--deploy-hash'), dict(action='append', help='Filter by (possibly multiple) deploy hash(es)')],
        [('--min-event-id',), dict(required=False, default=0, type=int, help="Supports replaying events from a given ID. If the value is 0, it it will subscribe to future events; if it's non-zero, it will replay all past events from that ID, without subscribing to new. To catch up with events from the beginning, start from 1.")],
    ])
    # fmt:on
    return parser.run([str(a) for a in arguments])


def main():
    return cli(*sys.argv[1:])


if __name__ == "__main__":
    sys.exit(main())



from collections import defaultdict

"""
Module for generating DAG diagrams in Graphviz DOT language.
"""

# Helpers for DOT language generation

INVISIBLE = "invis"  # "dashed" #"invis"


def attributes(**kwargs):
    if not kwargs:
        return ""
    return "[" + " ".join(f"{name}={value}" for (name, value) in kwargs.items()) + "]"


def node(node_id, **kwargs):
    return f'    "{node_id}" {attributes(**kwargs)}'


def edge(l1, l2, **kwargs):
    return f'    "{l1}" -> "{l2}" {attributes(**kwargs)}'


def subgraph(*args, **kwargs):
    style = kwargs.get("style", "") and f"""style="{kwargs['style']}" """
    label = kwargs.get("label", "") and f"""label="{kwargs['label']}" """
    return f"""  subgraph "{kwargs['name']}" {{
    {label}
    {style}
{cat(*args)}
  }}
"""


def graph(*args):
    return f"""
digraph "dag" {{
  rankdir=BT
  node [width=0 height=0 margin=0.03 fontsize=8 shape=box]
  splines=false
{cat(*args)}
}}
"""


def cat(*args):
    """
    Flatten lists and concatenate strings in them.
    """
    return "\n".join(type(arg) == list and cat(*arg) or str(arg) for arg in args)


# Protobuf BlockInfo abstraction/shortcuts.


def justifications(block_info):
    return [j.latest_block_hash.hex() for j in block_info.summary.header.justifications]


def rank(block_info):
    return block_info.summary.header.rank


def parents(block_info):
    ps = [h.hex() for h in block_info.summary.header.parent_hashes]
    if len(ps) == 0:
        # Genesis block doesn't have parents.
        return [""]
    return ps


def main_parent(block_info):
    return parents(block_info)[0]


def block_hash(block_info):
    return block_info.summary.block_hash.hex()


def short_hash(s: str) -> str:
    return f"{s[:10]}..."


def block_id(block_info) -> str:
    return short_hash(block_hash(block_info))


# DAG diagram construction.


def lane_alignment(block_infos, validator_id, min_rank, max_rank):
    """
    Create invisible edges between nodes in a lane to keep them rendered in order of their ranks.
    Create invisible nodes for missing ranks.
    """
    ranks = set(rank(b) for b in block_infos)
    edges = []
    nodes = []

    block_ids_by_rank = defaultdict(list)
    for b in block_infos:
        block_ids_by_rank[rank(b)].append(block_id(b))

    def dummy_id(i):
        return f"{i}_{validator_id}"

    for i in range(min_rank, max_rank):
        if i not in ranks and not block_ids_by_rank[i]:
            block_ids_by_rank[i] = [dummy_id(i)]
            nodes.append(node(block_ids_by_rank[i][0], style=INVISIBLE))

        if i + 1 not in ranks:
            block_ids_by_rank[i + 1] = [dummy_id(i + 1)]
            nodes.append(node(block_ids_by_rank[i + 1][0], style=INVISIBLE))

        edges.extend(
            [
                edge(n1, n2, style=INVISIBLE)
                for n1 in block_ids_by_rank[i]
                for n2 in block_ids_by_rank[i + 1]
            ]
        )

    return block_ids_by_rank, [nodes, edges]


def lane(validator, block_infos, min_rank, max_rank, genesis_block_id):
    validator_id = short_hash(validator)
    nodes = [node(block_id(b)) for b in block_infos]
    block_ids_by_rank, alignment_in_lane = lane_alignment(
        block_infos, validator_id, min_rank, max_rank
    )
    return subgraph(
        nodes,
        alignment_in_lane,
        name=f"cluster_{validator_id}",
        label=f"{validator_id}",
    )


def generate_dot(block_infos, show_justification_lines=False):
    genesis_block_id = "genesis_block"
    validator_blocks = defaultdict(list)
    for b in block_infos:
        pk = b.summary.header.validator_public_key
        if len(pk):
            validator_blocks[pk.hex()].append(b)
        else:
            genesis_block_id = block_id(b)

    block_hashes = set(block_hash(b) for b in block_infos)
    ranks = set(rank(b) for b in block_infos)
    min_rank = min(ranks)
    max_rank = max(ranks)
    lanes = [
        lane(validator, block_infos, min_rank, max_rank, genesis_block_id)
        for (validator, block_infos) in validator_blocks.items()
    ]

    # Showing only the first, main parent of each block.
    parent_edges = [
        [
            edge(
                block_id(b),
                short_hash(main_parent(b)),
                style="bold",
                constraint="false",
            )
            for b in block_infos
            if main_parent(b) in block_hashes
        ]
        for (validator, block_infos) in validator_blocks.items()
    ]

    lanes_alignment = [
        edge(genesis_block_id, alignment_node_id, style=INVISIBLE)
        for (validator, block_infos) in sorted(validator_blocks.items())
        for alignment_node_id in lane_alignment(
            block_infos, short_hash(validator), min_rank, max_rank
        )[0][min_rank]
    ]

    justification_lines = ""
    if show_justification_lines:
        justification_lines = [
            [
                edge(
                    block_id(b),
                    short_hash(j),
                    constraint="false",
                    style="dotted",
                    arrowhead="none",
                )
                for j in justifications(b)
                if j in block_hashes
            ]
            for b in block_infos
        ]

    genesis_block = (
        genesis_block_id != "genesis_block"
        and node(genesis_block_id)
        or node(genesis_block_id, style=INVISIBLE)
    )

    return graph(
        genesis_block, lanes_alignment, lanes, parent_edges, justification_lines
    )



# Generated by the Protocol Buffers compiler. DO NOT EDIT!
# source: diagnostics.proto
# plugin: grpclib.plugin.main
import abc
import typing

import grpclib.const
import grpclib.client
if typing.TYPE_CHECKING:
    import grpclib.server

from . import empty_pb2
from . import node_pb2
from . import diagnostics_pb2


class DiagnosticsBase(abc.ABC):

    @abc.abstractmethod
    async def ListPeers(self, stream: 'grpclib.server.Stream[empty_pb2.Empty, diagnostics_pb2.Peers]') -> None:
        pass

    def __mapping__(self) -> typing.Dict[str, grpclib.const.Handler]:
        return {
            '/io.casperlabs.node.api.diagnostics.Diagnostics/ListPeers': grpclib.const.Handler(
                self.ListPeers,
                grpclib.const.Cardinality.UNARY_UNARY,
                empty_pb2.Empty,
                diagnostics_pb2.Peers,
            ),
        }


class DiagnosticsStub:

    def __init__(self, channel: grpclib.client.Channel) -> None:
        self.ListPeers = grpclib.client.UnaryUnaryMethod(
            channel,
            '/io.casperlabs.node.api.diagnostics.Diagnostics/ListPeers',
            empty_pb2.Empty,
            diagnostics_pb2.Peers,
        )
